from pathlib import Path

from lxml import etree


def _get_omml2mml_transformation() -> etree.XSLT:
    omml2mml_file = Path(Path(__file__).parent, "xsl", "omml2mml", "OMML2MML.XSL")
    omml2mml = etree.XSLT(etree.parse(omml2mml_file))
    return omml2mml


def _get_mml2tex_transformation() -> etree.XSLT:
    mml2tex_file = Path(Path(__file__).parent, "xsl", "mml2tex", "mmltex.xsl")
    mml2tex = etree.XSLT(etree.parse(mml2tex_file))

    return mml2tex



from enum import Enum
from json import load
from typing import Iterator

from langchain_core.document_loaders import BaseLoader
from langchain_core.documents import Document
from langchain_core.load import load as ln_load

from protollm.raw_data_processing.docs_transformers import ChunkMerger, RecursiveSplitter
from protollm.raw_data_processing.docs_transformers.metadata_sentence_splitter import DivMetadataSentencesSplitter
from protollm.raw_data_processing.docs_transformers.key_words_splitter import MultiMetadataAppender


transformer_object_dict = {
    'recursive_character': RecursiveSplitter,
    # 'list_hierarchy': ListHierarchySplitter,
    'hierarchical_merger': ChunkMerger,
    'div_sentence_splitter': DivMetadataSentencesSplitter,
    'keyword_appender': MultiMetadataAppender
}


class LoaderType(str, Enum):
    docx = 'docx'
    doc = 'doc'
    odt = 'odt'
    rtf = 'rtf'
    pdf = 'pdf'
    directory = 'directory'
    zip = 'zip'
    json = 'json'


class LangChainDocumentLoader(BaseLoader):
    def __init__(self, file_path: str):
        self.file_path = file_path

    def lazy_load(self) -> Iterator[Document]:
        with open(self.file_path, 'r') as f:
            for i, doc_dict in load(f).items():
                yield ln_load(doc_dict)



class ConvertingError(Exception):
    def __init__(self, message):
        super().__init__(message)


class EncodingError(Exception):
    def __init__(self, message):
        super().__init__(message)


class NoTextLayerError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ChaptersExtractingFailedWarning(Warning):
    def __init__(self, message):
        super().__init__(message)


class ParseImageWarning(Warning):
    def __init__(self, message):
        super().__init__(message)


class TitleExtractingWarning(Warning):
    def __init__(self, message):
        super().__init__(message)


class PageNumbersExtractingWarning(Warning):
    def __init__(self, message):
        super().__init__(message)


class FooterExtractingWarning(Warning):
    def __init__(self, message):
        super().__init__(message)



import html
import re
from typing import Optional, Iterator
from uuid import uuid4

import docx
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.table import Table, _Cell, _Row
from docx.text.hyperlink import Hyperlink
from docx.text.paragraph import Paragraph
from ftfy import fix_text
from tabulate import tabulate

from protollm.raw_data_processing.docs_parsers.parsers.utilities import is_bulleted_text
from protollm.raw_data_processing.docs_parsers.parsers.word_doc.docx_parsing_config import (
    DocxParsingConfig,
)
from protollm.raw_data_processing.docs_parsers.parsers.word_doc.xml import process_paragraph_body


def _get_list_level(split_text: list[str], level: int = -1) -> int:
    if not re.search(r"[а-яА-ЯёЁ]", "".join(split_text[1:])):
        return level
    if not split_text:
        return level
    if not split_text[0].isdigit():
        return level
    else:
        level += 1
        return _get_list_level(split_text[1:], level)


def _get_urls(paragraph: Paragraph) -> dict[str, str]:
    urls = {}
    for item in paragraph.iter_inner_content():
        if isinstance(item, Hyperlink):
            text = item.text
            url = item.url
            if not text or not url:
                continue
            urls[text] = url
    return urls


def _get_metadata(paragraph: Optional[Paragraph] = None) -> dict:
    bold = False
    font_size = -1
    urls = {}
    list_level = _get_list_level([])
    is_bullet_list = False
    is_centered = False
    first_line_indent = 0

    if paragraph is not None:
        bold = paragraph.runs[0].bold or bold if paragraph.runs else bold
        bold = paragraph.style.font.bold or bold

        font_size = (
            paragraph.runs[0].font.size or font_size if paragraph.runs else font_size
        )

        urls.update(_get_urls(paragraph))

        split_text = re.split("[ .\xa0]", paragraph.text)
        list_level = _get_list_level(split_text)

        is_bullet_list = is_bulleted_text(paragraph.text) | (
            "<w:numPr>" in paragraph._element.xml
        )

        is_centered = paragraph.paragraph_format.alignment is WD_ALIGN_PARAGRAPH.CENTER

        paragraph_first_line_indent = paragraph.paragraph_format.first_line_indent
        first_line_indent = (
            paragraph_first_line_indent.emu
            if paragraph_first_line_indent is not None
            else first_line_indent
        )

    return {
        "bold": bold,
        "list_level": list_level,
        "is_bullet_list": is_bullet_list,
        "urls": urls,
        "is_centered": is_centered,
        "first_line_indent": first_line_indent,
        "font_size": font_size,
        "images": {},
        "formulas": {},
        "tables": {},
    }


def _process_paragraph(
    paragraph: Paragraph, parsing_config: DocxParsingConfig
) -> tuple[str, dict]:
    paragraph_text, paragraph_metadata = process_paragraph_body(
        paragraph, parsing_config=parsing_config
    )
    paragraph_text = fix_text(paragraph_text).replace("\xa0", " ")
    paragraph_text = " ".join(paragraph_text.split())

    metadata = _get_metadata(paragraph)
    metadata.update(paragraph_metadata)

    return paragraph_text, metadata


def _convert_to_html(
    table: Table, parsing_config: DocxParsingConfig, is_nested: bool = False
) -> tuple[str, list[dict]]:
    cells_metadata = []

    def iter_cell_block_items(cell: _Cell) -> Iterator[str]:
        for block_item in cell.iter_inner_content():
            if isinstance(block_item, Table):
                inner_html_table, inner_cells_metadata = _convert_to_html(
                    block_item, parsing_config, is_nested=True
                )
                cells_metadata.extend(inner_cells_metadata)
                yield inner_html_table
            elif isinstance(block_item, Paragraph):
                paragraph_text, metadata = _process_paragraph(
                    block_item, parsing_config
                )
                cells_metadata.append(metadata)
                yield f"{html.escape(paragraph_text)}"

    def iter_cells(row: _Row) -> Iterator[str]:
        return ("\n".join(iter_cell_block_items(cell)) for cell in row.cells)

    return (
        tabulate(
            [list(iter_cells(row)) for row in table.rows],
            headers=[] if is_nested else "firstrow",
            tablefmt="unsafehtml",
        ),
        cells_metadata,
    )


def _process_table(table: Table, parsing_config: DocxParsingConfig) -> tuple[str, dict]:
    html_table, cells_metadata = _convert_to_html(table, parsing_config)
    table_name = "table-" + str(uuid4())
    metadata = _get_metadata()
    metadata["tables"] = {table_name: html_table}
    for cell_meta in cells_metadata:
        for key in ["images", "formulas", "urls"]:
            metadata[key] = {**metadata.get(key, {}), **cell_meta.get(key, {})}

    return html_table, metadata  # TODO: f'{{{table_name}}}' instead of html_table


def parse_docx_to_lines(
    stream,
    extract_tables: bool = False,
    extract_images: bool = False,
    extract_formulas: bool = False,
) -> tuple[list[str], list[dict]]:
    lines, metadata = [], []
    document = docx.Document(stream)
    parsing_config = DocxParsingConfig(document, extract_images, extract_formulas)
    for section in document.sections:
        for block_item in section.iter_inner_content():
            if isinstance(block_item, Paragraph):
                line, meta = _process_paragraph(block_item, parsing_config)
            elif isinstance(block_item, Table) and extract_tables:
                line, meta = _process_table(block_item, parsing_config)
            else:
                continue

            if not line:
                continue

            lines.append(line)
            metadata.append(meta)

    return lines, metadata



import re


def fix_list_dots_separators(sentences: list[str]) -> list[str]:
    """
    Takes list of sentences and combines those of them that are list items
    that were incorrectly separated due to the use of a dot separator.
    Returns updated list of sentences
    """
    fixed_sentences_lst = []
    sentence_parts_lst = []
    i = 0
    while i < len(sentences):
        chunk = sentences[i].strip()
        if len(chunk) > 0:
            # it means that the dot was used to separate list elements, and we should join such sentences
            if not chunk[0].isupper() and not chunk[0].isdigit():
                sentence_parts_lst.append(chunk)
            else:
                if len(sentence_parts_lst) != 0:
                    fixed_sentences_lst.append("; ".join(sentence_parts_lst))
                sentence_parts_lst = [chunk]
        i += 1

    if len(sentence_parts_lst) != 0:
        fixed_sentences_lst.append("; ".join(sentence_parts_lst))
    return fixed_sentences_lst




# Пример заполнения шаблона - objects = 'характеристики различных зданий в Санкт-Петербурге',
# user_field = 'ГИС-системы', N=10

synthetic_prompt_template_basic = '''В файле даны {objects}. 
Проанализируй файл и сформулируй {N} вопросов, которые мог бы задать 
пользователь {user_field}.'''

# data_desc = 'зданий в Санкт-Петербурге', stucture - JSON-описание, role - 'урбанист',
# work = 'анализ различных  данных о городе для получения некоторых практических выводов'
synthetic_prompt_template_advanced = ''' 
Датасет с данными {data_desc} имеет следующую структуру: {structures}. 
Представь, что ты - {role}. В круг твоих задач входит {work}. 
Сформулируй {N} вопросов, которые мог бы задать {role}, основываясь на 
данных датасета.'''

prompt_template_enrichment = """Теперь представь, что у тебя есть другие данные о {data_desc}, и 
можно ориентироваться не только на структуру датасета. Сгенерируй ещё {N} вопросов по теме {data_desc}."""

prompt_template_file_only = """В файле даны характеристики различных {data_desc}. 
Проанализируй файл и сформулируй {N} вопросов, которые мог бы 
задать пользователь {user_field}. Ответь на них на основе информации из файла."""

prompt_template_expert_role = """Представь, что ты - {role}, 
который занимается {work}. Вот примеры вопросов, которые задают специалисты в этой области: 
{questions}. Сформулируй ещё {N} вопросов. Они могут быть как широкими и обобщёнными, так и узкоспециализированными.
"""



from protollm.rags.settings.pipeline_settings import PipelineSettings
from protollm.rags.settings.es_settings import settings as es_settings
from protollm.rags.settings.chroma_settings import settings as default_settings



from enum import Enum


class DocType(str, Enum):
    docx = "docx"
    doc = "doc"
    odt = "odt"
    rtf = "rtf"
    pdf = "pdf"
    zip = "zip"
    unsupported = "unsupported"  # TODO: add txt, xlsx, pptx support


class ConvertingDocType(str, Enum):
    docx = "docx"  # TODO: add xlsx, pptx support


class ParsingScheme(str, Enum):
    paragraphs = "paragraphs"
    lines = "lines"
    chapters = "chapters"
    full = "full"



BASE_PROMPT = """You are a good assistant, who will be offered $100 tips for each correct answer."""

FC_SYSTEM_PROMPT = """You are a helpful assistant with access to the following functions. Use them if required - ${tools}."""

FC_USER_PROMPT = """Extract all relevant data for answering this question: ${question}. You MUST return ONLY the function names separated by spaces.
    Do NOT return any other additional text."""

CLASSFIC_FC_PROMPT = """You have access to the following functions:
    Use the function '{function_name]}' to '{function_description}':
    {function_json}\\
    If you choose to call a function ONLY, reply in the following 
    format with no prefix or suffix:
    <function=example_function_name>
    {{'example_name': 'example_value'}}
    </function>
    Reminder:
     - Function calls MUST follow the specified format, 
       start with <function= and end with </function>
     - Required parameters MUST be specified
     - Only call one function at a time
     - Put the entire function call reply on one line
     - If there is no function call available, answer the question 
       like normal with your current knowledge and do not tell 
       the user about function calls"""

EXT_SERVICE_PROMPT="""    Answer the question by following the rules below.
    For the answer you must use the context provided by user.
    Rules:
    1. You must only use provided information for the answer.
    2. Add a unit of measurement to the answer.
    3. For answer you should take only the information from context, 
    which is relevant to the user's question.
    4. If an interpretation is provided in the context 
    for the data requested in the question,
    it should be added in the answer.
    5. If data for an answer is absent, reply that data 
    was not provided or absent and
    mention for what field there was no data.
    6. If you do not know how to answer the questions, say so.
    7. Before giving an answer to the user question, 
    provide an explanation. Mark the answer
    with keyword 'ANSWER', and explanation with 'EXPLANATION'.
    8. If the question is about complaints, 
    answer about at least 5 complaints topics.
    9. Answer should be three sentences maximum.
"""

RAG_QA_PROMPT="""
    Answer the question following the rules below. For answer 
    you must use context provided by the user.
    Rules:
    1. You must use only provided information for the answer.
    2. For answer you should take only that information 
    from context, which is relevant to
    user's question.
    3. If data for an answer is absent, answer that 
    data was not provided or absent and
    mention for what field there was no data.
    4. If you do not know how to answer the questions, say so.
    5. Before giving an answer to the user question, 
    provide an explanation. Mark the answer
    with keyword 'ANSWER', and explanation with 'EXPLANATION'.
    6. The answer should consist of as many sentences 
    as are necessary to answer the
    question given the context, but not more five sentences.
"""


import requests
import json

from langchain.agents import create_structured_chat_agent, AgentExecutor
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.prompts import SystemMessagePromptTemplate, HumanMessagePromptTemplate
import pandas as pd
from langchain_core.tools import tool

from protollm.agents.llama31_agents.llama31_agent import Llama31ChatModel
from examples.real_world.chemical_pipeline.validate_tools import validate_decompose, compute_metrics, validate_conductor

# Create the system and human prompts
system_prompt = '''
Respond to the human as helpfully and accurately as possible.
'''

human_prompt = '''{input}
{agent_scratchpad}
(Reminder to respond in a JSON blob no matter what)'''

system_message = SystemMessagePromptTemplate.from_template(
    system_prompt,
    input_variables=["tools", "tool_names"],
)
human_message = HumanMessagePromptTemplate.from_template(
    human_prompt,
    input_variables=["input", "agent_scratchpad"],
)

# Create the ChatPromptTemplate
prompt = ChatPromptTemplate.from_messages(
    [
        system_message,
        MessagesPlaceholder(variable_name="chat_history", optional=True),
        human_message,
    ]
)

# Initialize the custom LLM
llm = Llama31ChatModel(
    api_key='API_KEY_HERE',
    base_url="URL_HERE",
    model="meta-llama/llama-3.1-70b-instruct",
    temperature=0.5,
    max_tokens=5000
)

# Create the structured chat agent
agent = create_structured_chat_agent(
    llm=llm,
    prompt=prompt,
    stop_sequence=True
)

# Create the AgentExecutor
agent_executor = AgentExecutor.from_agent_and_tools(
    verbose=True,
    return_intermediate_steps=True,
    output_keys=["output"],
    early_stopping_method="generate"
)

response = agent_executor.invoke({
    "input": 'test'
})



from os.path import dirname
from pathlib import Path
from pydantic_settings import BaseSettings, SettingsConfigDict


class ChromaSettings(BaseSettings):
    # Chroma DB settings
    chroma_host: str = 'any'
    chroma_port: int = 8888
    allow_reset: bool = False

    # Documents collection's settings
    collection_name: str = 'collection'
    collection_names_for_advance: list[str] = ['collection']
    embedding_name: str = 'intfloat/multilingual-e5-large'
    embedding_host: str = ''
    distance_fn: str = 'cosine'

    # Documents' processing settings
    docs_processing_config: str = str(Path(dirname(dirname(__file__)), '/config_files/', 'docs_processing_config.yaml'))
    docs_collection_path: str = str(Path(dirname(dirname(dirname(__file__))), '/docs/', 'example.docx'))

    model_config = SettingsConfigDict(
        env_file=Path(dirname(dirname(__file__)), '/config_files/', 'chroma.env'),
        env_file_encoding='utf-8',
        extra='ignore',
    )


settings = ChromaSettings()



import uuid
from collections import defaultdict
from typing import Any, Iterable, Callable

import chromadb
import numpy as np
from langchain_community.vectorstores.chroma import Chroma
from langchain_core.documents import Document
from tqdm import tqdm


def merge_collections(chroma_client: chromadb.HttpClient,
                      collection_name_1: str,
                      collection_name_2: str,
                      new_collection_name: str | None = None):
    """
    Merge 2 collections into the 'collection_name_1' if 'new_collection_name' is None,
    otherwise merge 2 collections into the new one with 'new_collection_name' name.
    If any problems with network or DB accessibility will occur, exception are raised

    :raises Exception: if there are network or database accessibility issues.
    """

    collection_1 = chroma_client.get_collection(name=collection_name_1)
    collection_2 = chroma_client.get_collection(name=collection_name_2)

    docs_1: dict[str, Any] = collection_1.get(include=['documents', 'metadatas', 'embeddings'])
    docs_2: dict[str, Any] = collection_2.get(include=['documents', 'metadatas', 'embeddings'])

    if new_collection_name is None:
        for i in range(len(docs_2['ids'])):
            if docs_2['documents'][i] not in docs_1['documents']:
                collection_1.add(ids=[str(uuid.uuid4())], metadatas=[docs_2['metadatas'][i]], documents=[docs_2['documents'][i]], embeddings=[docs_2['embeddings'][i]])
        return

    new_collection = chroma_client.create_collection(name=new_collection_name)

    merged_docs = docs_1
    for i in range(len(docs_2['ids'])):
        if docs_2['documents'][i] not in merged_docs['documents']:
            merged_docs['ids'].append(str(uuid.uuid4()))
            merged_docs['embeddings'].append(docs_2['embeddings'][i])
            merged_docs['metadatas'].append(docs_2['metadatas'][i])
            merged_docs['documents'].append(docs_2['documents'][i])

    for i in range(len(merged_docs['ids'])):
        new_collection.add(ids=[str(uuid.uuid4())], metadatas=merged_docs['metadatas'][i], documents=merged_docs['documents'][i])


def delete_repeats(collection: Chroma, similarity_func: Callable[[list[float], list[float]], float] = None) -> None:
    """
    Remove duplicate documents from a collection.

    :param collection: it should include fields: 'documents', 'embeddings', 'metadatas'

    :raises Exception: if there are issues accessing the database.
    """
    def cosine_similarity(vector1: list[float], vector2: list[float]) -> float:
        return np.dot(vector1, vector2) / (np.linalg.norm(vector1) * np.linalg.norm(vector2))

    if similarity_func is None:
        similarity_func = cosine_similarity
    docs = collection.get(include=['documents', 'metadatas', 'embeddings'])
    ids = docs['ids']
    documents = docs['documents']
    embeddings = docs['embeddings']
    delete_ids = []
    cache_ids = defaultdict(list)

    for i in tqdm(range(len(docs['ids']))):
        for j in range(i):
            if j in cache_ids[i]:
                continue
            if similarity_func(embeddings[i], embeddings[j]) > 0.97:
                if cache_ids[j]:
                    cache_ids[i].extend(cache_ids[j])
                    break
                cache_ids[i].append(j)
        delete_ids.extend([ids[j] if len(documents[j]) > len(documents[i]) else ids[i] for j in cache_ids[i]])

    return cache_ids
    # delete_ids = list(set(delete_ids))
    # collection.delete(ids=delete_ids)


def list_collections(client: chromadb.HttpClient) -> Iterable[str]:
    return [collection.name for collection in client.list_collections()]


def get_all_docs_name(collection: Chroma) -> set[str]:
    """
    Return list of files' name from collection.

    :param collection: it should include fields: 'documents', 'embeddings', 'metadatas'

    :raises KeyError: if there is no key 'source' in the documents' metadata from 'collection'
    """
    docs: dict[str, Any] = collection.get()

    if 'source' not in docs['metadatas'][0].keys():
        raise KeyError('There is no file name, called <source>, in document metadata')

    return set(str(metadata['source'].split('\\')[-1]) for metadata in docs['metadatas'])


def insert_documents(collection: Chroma, docs: Iterable[Document]):
    """
    Insert only documents whose names aren't in 'collection'

    :raises KeyError: if there is no key 'source' in the documents' metadata from 'collection' or 'docs'
    """
    first_element = next(docs)
    if 'source' not in first_element.metadata.keys():
        raise KeyError('There is no file name, called <source>, in document metadata')

    existing_docs_name = set(get_all_docs_name(collection))
    new_docs_name = set([str(doc.metadata['source'].split('\\')[-1]) for doc in docs] +
                        [str(first_element.metadata['source'].split('\\')[-1])])
    docs_name_for_insert = new_docs_name.difference(existing_docs_name)
    if first_element.metadata['source'].split('\\')[-1] in docs_name_for_insert:
        docs_for_insert = [first_element]
    else:
        docs_for_insert = []
    docs_for_insert += [doc for doc in docs if doc.metadata['source'].split('\\')[-1] in docs_name_for_insert]
    if docs_for_insert:
        collection.add_documents(docs_for_insert)




import logging
from typing import Any, Iterable, List, Optional

from langchain_text_splitters.character import RecursiveCharacterTextSplitter

logger = logging.getLogger(__name__)


class RecursiveSplitter(RecursiveCharacterTextSplitter):
    """Splitting text by the given sequence of splitters.
    """

    def __init__(
        self,
        separators: Optional[List[str]] = None,
        keep_separator: bool = True,
        is_separator_regex: bool = False,
        **kwargs: Any,
    ) -> None:
        """Create a new TextSplitter."""
        kwargs["chunk_overlap"] = 0
        super().__init__(keep_separator=keep_separator, **kwargs)
        self._separators = separators or [
            "\n\n",
            "\n",
            ". ",
            ";",
            ", ",
            ".",
            ",",
            " ",
            "",
        ]
        self._is_separator_regex = is_separator_regex

    def split_text(self, text: str) -> List[str]:
        if self._length_function(text) < self._chunk_size:
            return [text]
        return self._split_text(text, self._separators)

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        docs = []
        current_doc = []
        for text in splits:
            merged_text = self._join_docs([*current_doc, text], separator)
            if (
                merged_text is None
                or self._length_function(merged_text) <= self._chunk_size
            ):
                current_doc.append(text)
                continue
            doc = self._join_docs(current_doc, separator)
            if doc is None and self._length_function(text) > self._chunk_size:
                logger.warning(
                    f"Created a chunk, which is longer than the specified {self._chunk_size}"
                )
            else:
                docs.append(doc)
            current_doc = [text]
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs



import os
import subprocess
from pathlib import Path
from typing import Union

from protollm.raw_data_processing.docs_parsers.utils.exceptions import ConvertingError


def _convert_with_soffice(
    filename: Union[Path, str],
    output_directory: Union[Path, str],
    target_doc_type: str = "docx",
    timeout: int = None,
):
    """
    Converts a file to a target format using the libreoffice CLI.
    """
    command = [
        "soffice",
        "--headless",
        "--convert-to",
        target_doc_type,
        "--outdir",
        output_directory,
        str(filename),
    ]
    expected_path = Path(
        output_directory, ".".join((Path(filename).stem, target_doc_type))
    )
    try:
        result = subprocess.run(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=timeout
        )
        try:
            error = result.stderr.decode().strip()
        except UnicodeError:
            error = "*** Error message cannot be displayed ***"
        if error and not os.path.isfile(expected_path):
            raise ConvertingError(
                f"Could not convert file to {target_doc_type}\n{error}"
            )
    except subprocess.TimeoutExpired:
        raise ConvertingError(
            f"Converting file to {target_doc_type} hadn't terminated after {timeout} seconds"
        ) from None
    except FileNotFoundError:
        raise ConvertingError(
            "soffice command was not found. Please install libreoffice on your system and try again."
        ) from None



from functools import partial

from protollm.raw_data_processing.docs_parsers.parsers.converting.converted_file import (
    converted_file as __converted_file,
)
from protollm.raw_data_processing.docs_parsers.parsers.entities import ConvertingDocType

converted_file_to_docx = partial(
    __converted_file, target_doc_type=ConvertingDocType.docx
)



from copy import deepcopy

import chromadb
from langchain_community.embeddings.huggingface_hub import HuggingFaceHubEmbeddings
from langchain_community.vectorstores.chroma import Chroma

from protollm.rags.rag_core.retriever import DocRetriever, DocsSearcherModels
from protollm.rags.settings.chroma_settings import ChromaSettings, settings as default_settings
from protollm.rags.stores.chroma.chroma_loader import load_documents_to_chroma_db


def chroma_loading(path: str, collection: str) -> None:
    # Loads data to ChromaDB

    default_settings.collection_name = collection
    default_settings.docs_collection_path = path
    processing_batch_size = 32
    loading_batch_size = 32
    settings = deepcopy(default_settings)
    load_documents_to_chroma_db(
        settings=settings,
        processing_batch_size=processing_batch_size,
        loading_batch_size=loading_batch_size,
    )


def chroma_view(
    query: str,
    collection: str,
    k: int = 1,
    embedding_function: HuggingFaceHubEmbeddings = None,
) -> list:
    # Returns k chunks that are closest to the query
    chroma_client = init_chroma_client()
    embedding_function = HuggingFaceHubEmbeddings(model=default_settings.embedding_host)
    default_settings.collection_name = collection

    chroma_collection = Chroma(
        collection_name=collection,
        embedding_function=embedding_function,
        client=chroma_client,
    )

    return chroma_collection.similarity_search_with_score(query, k)


def init_chroma_client():
    return chromadb.HttpClient(
        host=default_settings.chroma_host,
        port=default_settings.chroma_port,
        settings=chromadb.Settings(allow_reset=default_settings.allow_reset),
    )


def proto_view(
    query: str,
    collection: str,
    k: int = 1,
) -> list:
    # Returns k chunks that are closest to the query
    embedding_function = HuggingFaceHubEmbeddings(model=default_settings.embedding_host)
    chroma_client = init_chroma_client()

    docs_searcher_models = DocsSearcherModels(embedding_model=embedding_function, chroma_client=chroma_client)
    retriever = DocRetriever(top_k=k,
                             docs_searcher_models=docs_searcher_models,
                             )

    return retriever.retrieve_top(collection_name=collection, query=query)


def delete_collection(collection: str) -> None:
    # Deletes the collection
    chroma_client = init_chroma_client()
    chroma_client.delete_collection(collection)


def list_collections() -> list:
    # Returns a list of all collections
    chroma_client = init_chroma_client()
    return chroma_client.list_collections()


def create_base_collection(
    collection_name: str,
    docs_collection: list[str],
) -> chromadb.Collection:
    """Creates a new collection in the ChromaDB and adds the given documents to it.

    Args:
        collection_name (str): The name of the new collection.
        docs_collection (list[str]): The list of documents to add to the collection.
        Must be just a list of any strings, which should not be too long due to vectorization.

    Returns:
        chromadb.Collection: The newly created collection.
    """
    chroma_client = init_chroma_client()
    embedding_function = HuggingFaceHubEmbeddings(model=default_settings.embedding_host)
    collection = chroma_client.create_collection(
        collection_name,
        embedding_function=embedding_function,
    )
    collection.add(
        documents=docs_collection,
        ids=[str(i) for i in range(len(docs_collection))],
    )
    return collection


if __name__ == "__main__":
    collection_name = "external-document"
    query = "Как называется документ?"
    res = proto_view(query, collection_name)
    print(res[0][0].page_content)



from docx.document import Document

from protollm.raw_data_processing.docs_parsers.parsers.word_doc.xml.utilities import (
    _get_omml2mml_transformation,
    _get_mml2tex_transformation,
)


class DocxParsingConfig:
    def __init__(
        self,
        document: Document,
        extract_images: bool = False,
        parse_formulas: bool = False,
    ):
        self.__rels = document.part.rels
        self.__extract_images = extract_images
        self.__parse_formulas = parse_formulas
        self.__omml2mml = None
        self.__mml2tex = None

    @property
    def extract_images(self):
        return self.__extract_images

    @property
    def parse_formulas(self):
        return self.__parse_formulas

    @property
    def document_relationships(self):
        return self.__rels

    @property
    def omml2mml_transformation(self):
        if self.__omml2mml is None and self.__parse_formulas:
            self.__omml2mml = _get_omml2mml_transformation()

        return self.__omml2mml

    @property
    def mml2tex_transformation(self):
        if self.__mml2tex is None and self.__parse_formulas:
            self.__mml2tex = _get_mml2tex_transformation()

        return self.__mml2tex



from contextlib import contextmanager
from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import BinaryIO, Generator, Union, Optional

from protollm.raw_data_processing.docs_parsers.parsers.converting.converting import _convert_with_soffice
from protollm.raw_data_processing.docs_parsers.parsers.entities import ConvertingDocType


@contextmanager
def converted_file(
    stream,
    target_doc_type: Union[str, ConvertingDocType] = ConvertingDocType.docx,
    timeout: Optional[int] = None,
) -> Generator[BinaryIO, None, None]:
    if target_doc_type not in ConvertingDocType.__members__:
        raise ValueError("Invalid target document type")
    target_doc_type = f"{target_doc_type}"

    with TemporaryDirectory() as tmp_dir:
        tmp_file = NamedTemporaryFile(delete=False, dir=tmp_dir)
        tmp_file.write(stream.read())
        tmp_file.close()
        tmp_file_path = tmp_file.name

        _convert_with_soffice(
            filename=tmp_file_path,
            output_directory=tmp_dir,
            target_doc_type=target_doc_type,
            timeout=timeout,
        )
        converted_file_path = ".".join((tmp_file_path, target_doc_type))

        with open(converted_file_path, "rb") as f:
            yield f



from typing import Union

from langchain_elasticsearch._utilities import DistanceStrategy
from langchain_elasticsearch.vectorstores import BaseRetrievalStrategy

from protollm.rags.settings import es_settings


class BM25RetrievalStrategy(BaseRetrievalStrategy):
    """BM25 retrieval strategy for ElasticSearch"""

    def index(
            self,
            dims_length: Union[int, None],
            vector_query_field: str,
            similarity: Union[DistanceStrategy, None],
    ) -> dict:
        index_kwargs = {"settings": es_settings.es_index_settings,
                        "mappings": es_settings.es_index_mappings}
        return index_kwargs

    def query(
            self,
            query_vector: Union[list[float], None],
            query: Union[str, None],
            *,
            k: int,
            fetch_k: int,
            vector_query_field: str,
            text_field: str,
            filter: list[dict],
            similarity: Union[DistanceStrategy, None],
    ) -> dict:
        if query is None:
            raise ValueError(
                "You must provide a query to perform a similarity search."
            )
        new_query = dict(es_settings.es_query_template)
        new_query['multi_match']['query'] = query
        query_body = {'query': new_query}
        return query_body

    def require_inference(self) -> bool:
        return False



import warnings
from typing import Any, Optional, Callable, Dict

from langchain_core.documents import Document

from langchain_chroma import Chroma

from dataclasses import dataclass

from langchain_community.embeddings.sentence_transformer import SentenceTransformerEmbeddings
from chromadb import ClientAPI


@dataclass
class DocsSearcherModels:
    embedding_model: SentenceTransformerEmbeddings | None | Any = None
    chroma_client: ClientAPI | None = None


class DocRetriever:
    def __init__(self, top_k: int, docs_searcher_models: DocsSearcherModels,
                 preprocess_query: Optional[Callable[[str], str]] = None):
        """
        :param preprocess_query: for example, get keywords from query
        """
        self.top_k = top_k
        self.embedding_function = docs_searcher_models.embedding_model
        self.client = docs_searcher_models.chroma_client
        self.preprocess_query = preprocess_query

    def retrieve_top(self, collection_name: str, query: str, filter: Optional[Dict[str, Any]] = None) \
            -> Optional[list[Document]]:
        """
        Retrieve top K documents from Vector DB (e.x., Chroma).
        """
        if filter is None:
            _filter = {}
        else:
            _filter = filter.copy()
        if collection_name is None:
            warnings.warn('Collection name is None')
            return None
        if self.preprocess_query is not None:
            query = self.preprocess_query(query)
        if collection_name not in [col.name for col in self.client.list_collections()]:
            warnings.warn('There is no collection named {} in Chroma DB'.format(collection_name))
            return None

        store = Chroma(
            client=self.client,
            collection_name=collection_name,
            embedding_function=self.embedding_function,
        )
        return store.as_retriever(search_kwargs={'k': self.top_k, 'filter': _filter}).invoke(query)


class RetrievingPipeline:
    def __init__(self):
        self._retrievers: Optional[list[DocRetriever]] = None
        self._collection_names: Optional[list[str]] = None

    def set_retrievers(self, retrievers: list[DocRetriever]) -> 'RetrievingPipeline':
        self._retrievers = retrievers
        return self

    def set_collection_names(self, collection_names: list[str]) -> 'RetrievingPipeline':
        self._collection_names = collection_names
        return self

    def get_retrieved_docs(self, query: str) -> list[Document]:
        if any([self._retrievers is None, self._collection_names is None]):
            raise ValueError('Either retrievers or collection_names must not be None')

        if len(self._retrievers) == len(self._collection_names):
            _query = query
            docs = self._retrievers[0].retrieve_top(self._collection_names[0], _query)
            for i in range(1, len(self._retrievers)):
                filter = {'uuid': {'$in': [doc.metadata['uuid'] for doc in docs]}}
                docs_next = self._retrievers[i].retrieve_top(self._collection_names[i], _query, filter)
                docs = docs_next
        else:
            raise Exception('The length of retrievers and collection_names must match')

        return docs



import logging
import uuid
from typing import Optional, List, Any, Callable

import chromadb
from langchain_community.embeddings.huggingface_hub import HuggingFaceHubEmbeddings
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models import LLM

from protollm_sdk.jobs.job import Job
from protollm_sdk.jobs.job_context import JobContext
from protollm_sdk.models.job_context_models import PromptModel, PromptMeta, ResponseModel

from protollm.templates.prompt_templates.rag_prompt_templates import EOS, EOT
from protollm.rags.rag_core.retriever import DocsSearcherModels, DocRetriever, RetrievingPipeline
from protollm.rags.rag_core.utils import run_multiple_rag
from protollm.rags.settings.chroma_settings import settings

from protollm.raw_data_processing.docs_transformers.key_words_splitter import KeywordExtractor

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def get_simple_retriever(docs_searcher_models: DocsSearcherModels, top_k: int,
                         preprocess_query: Optional[Callable[[str], str]] = None) -> DocRetriever:
    return DocRetriever(top_k, docs_searcher_models, preprocess_query)


def get_advanced_retriever(docs_searcher_models: DocsSearcherModels, top_ks: list[int] = None):
    if top_ks is None:
        top_ks = [100, 50, 10]
    file_name_retriever = get_simple_retriever(docs_searcher_models, top_ks[0])
    keywords_retriever = get_simple_retriever(docs_searcher_models, top_ks[1],
                                              lambda query: KeywordExtractor().get_object_action_pair(query)[0])
    content_retriever = get_simple_retriever(docs_searcher_models, top_ks[2])
    return [file_name_retriever, keywords_retriever, content_retriever]


def get_pipelines_retriever(list_retrievers: list[list[DocRetriever]], list_collection_names: list[list[str]]) \
        -> list[RetrievingPipeline]:
    return [RetrievingPipeline().set_retrievers(retrievers).set_collection_names(collection_names)
            for retrievers, collection_names in zip(list_retrievers, list_collection_names)]


class CustomLLM(LLM):
    ctx: JobContext = None

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        temperature = kwargs.get("temperature", 0)
        tokens_limit = kwargs.get("tokens_limit", 8000)
        stop_words = kwargs.get("stop_words", [EOS, EOT])

        return self.ctx.outer_llm_api.inference(
            PromptModel(
                job_id=str(uuid.uuid4()),
                content=prompt,
                meta=PromptMeta(temperature=temperature, tokens_limit=tokens_limit, stop_words=stop_words)
            )
        ).content

    def _llm_type(self) -> str:
        return 'custom'


class RAGJob(Job):
    def run(self, job_id: str, ctx: JobContext, **kwargs: Any):
        """
        Run RAG pipeline and save the LLM response to outer database

        :param user_prompt: input question
        :param use_advanced_rag:
        :return: None
        """
        user_prompt: str | list[str] = kwargs['user_prompt']
        use_advanced_rag: bool = kwargs.get('use_advanced_rag', False)
        client = chromadb.HttpClient(host=settings.chroma_host, port=settings.chroma_port)
        llm = CustomLLM(ctx=ctx, n=10)
        encoder = HuggingFaceHubEmbeddings(huggingfacehub_api_token='any-token', model=settings.embedding_host)
        docs_searcher_models = DocsSearcherModels(embedding_model=encoder, chroma_client=client)

        logger.info('Necessary setting setup')

        simple_pipeline = get_pipelines_retriever([[get_simple_retriever(docs_searcher_models, 10)]], [[settings.collection_name]])
        advanced_pipeline = get_pipelines_retriever(
            [[get_simple_retriever(docs_searcher_models, 10)], get_advanced_retriever(docs_searcher_models)],
            [[settings.collection_name], settings.collection_names_for_advance]
        )

        pipelines = simple_pipeline if not use_advanced_rag else advanced_pipeline

        # Run RAG process and obtain the final LLM response
        llm_response = run_multiple_rag(user_prompt, llm, pipelines)
        logger.info(f'LLM responses the follow text: "{llm_response}"')

        # Save result to DB
        ctx.result_storage.save_dict(job_id, ResponseModel(content=llm_response).model_dump())



import os
import json
from pathlib import Path


CONFIG_PATH = Path(Path(__file__).parent, 'configs')


class ElasticsearchSettings:
    es_host: str = os.environ.get("ELASTIC_HOST", "")
    es_port: int = os.environ.get("ELASTIC_PORT", 80)
    es_url: str = f"http://{es_host}:{es_port}"
    es_user: str = os.environ.get("ELASTIC_USER", "")
    es_password: str = os.environ.get("ELASTIC_PASSWORD", "")

    es_index_mappings: dict = json.loads(Path(CONFIG_PATH, 'index_mappings.json').read_text(encoding="utf-8"))
    es_index_settings: dict = json.loads(Path(CONFIG_PATH, 'index_settings.json').read_text(encoding="utf-8"))
    es_query_template: dict = json.loads(Path(CONFIG_PATH, 'query_template.json').read_text(encoding="utf-8"))
    es_query_all_hits: dict = json.loads(Path(CONFIG_PATH, 'query_all_hits.json').read_text(encoding="utf-8"))

    metadata_fields: list[str] = list(es_index_mappings['properties']['metadata']['properties'].keys())
    content_field: str = 'paragraph'


settings = ElasticsearchSettings()



from protollm.raw_data_processing.docs_parsers.parsers.base import BaseParser
from protollm.raw_data_processing.docs_parsers.parsers.entities import DocType, ParsingScheme
from protollm.raw_data_processing.docs_parsers.parsers.pdf import PDFParser
from protollm.raw_data_processing.docs_parsers.parsers.word_doc import WordDocumentParser



import logging
from itertools import islice
from pathlib import Path
from typing import Iterable, Optional, Union

from langchain_community.vectorstores import utils
from langchain_core.documents import Document
from langchain_core.vectorstores import VectorStore

from protollm.rags.pipeline.docs_processing.utils import get_loader

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


class DocsExtractPipeline:
    def __init__(self, pipeline_settings: 'PipelineSettings'):
        self._pipeline_settings = pipeline_settings

    def update_loader(self, **kwargs) -> 'DocsExtractPipeline':
        self._pipeline_settings.update_loader_params(kwargs)
        return self

    def load_docs(self, docs_collection_path: Optional[Union[str, Path]] = None,
                  byte_content: Optional[bytes] = None):
        logger.info('Initialize parsing process')
        if docs_collection_path is not None:
            self.update_loader(file_path=docs_collection_path)

        loader = get_loader(byte_content=byte_content, **self._pipeline_settings.loader_params)
        return loader.lazy_load()

    def go_to_next_step(self, docs_collection_path: Optional[Union[str, Path]] = None,
                        byte_content: Optional[bytes] = None) -> 'DocsTransformPipeline':
        return DocsTransformPipeline(self._pipeline_settings, self.load_docs(docs_collection_path, byte_content))


class DocsTransformPipeline:
    def __init__(self, pipeline_settings: 'PipelineSettings', docs_generator: Iterable[Document]):
        self._pipeline_settings = pipeline_settings
        self._docs_generator = docs_generator

    def update_docs_transformers(self, **kwargs) -> 'DocsTransformPipeline':
        self._pipeline_settings.update_transformer_params(kwargs)
        return self

    def transform(self, batch_size: int = 100) -> Iterable[Document]:
        logger.info('Initialize transformation process')
        transformers = self._pipeline_settings.transformers
        batch_size = max(batch_size, 1)
        while docs_batch := list(islice(self._docs_generator, batch_size)):
            for transformer in transformers:
                docs_batch = transformer.transform_documents(docs_batch)
            yield from docs_batch

    def go_to_next_step(self, batch_size: int = 100) -> 'DocsLoadPipeline':
        return DocsLoadPipeline(self.transform(batch_size))


class DocsLoadPipeline:
    def __init__(self, docs_generator: Iterable[Document]):
        self._docs_generator = docs_generator

    def load(self, store: VectorStore, loading_batch_size: int = 32) -> None:
        logger.info('Initialize loading process')

        # Add processed documents to the store
        loading_batch_size = max(loading_batch_size, 1)
        while docs_batch := list(islice(self._docs_generator, loading_batch_size)):
            # TODO: replace filtering. It loses important metadatas. Use DocsTransformer to transform metadata
            store.add_documents(utils.filter_complex_metadata(docs_batch))




import zipfile
from pathlib import Path
from typing import Iterator, Union, Any, Optional, Sequence

from langchain_core.document_loaders import BaseLoader, Blob
from langchain_core.documents import Document
from tqdm import tqdm

from protollm.raw_data_processing.docs_parsers.parsers import (
    ParsingScheme,
    DocType,
    BaseParser,
    PDFParser,
    WordDocumentParser,
)
from protollm.raw_data_processing.docs_parsers.utils.logger import ParsingLogger
from protollm.raw_data_processing.docs_parsers.utils.utilities import correct_path_encoding


class ZipLoader(BaseLoader):
    """
    Load files from zip archive into list of documents.
    """

    def __init__(
        self,
        file_path: Union[str, Path],
        byte_content: Optional[bytes] = None,
        pdf_parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        pdf_extract_images: bool = False,
        pdf_extract_tables: bool = False,
        pdf_extract_formulas: bool = False,
        pdf_remove_service_info: bool = False,
        word_doc_parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        word_doc_extract_images: bool = False,
        word_doc_extract_tables: bool = False,
        word_doc_extract_formulas: bool = False,
        timeout_for_converting: Optional[int] = None,
        exclude_files: Sequence[Union[Path, str]] = (),
        parsing_logger: Optional[ParsingLogger] = None,
        silent_errors: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize with a file path."""
        self.file_path = str(file_path)
        doc_type = BaseParser.get_doc_type(self.file_path)
        if doc_type is not DocType.zip:
            if doc_type is DocType.unsupported:
                raise ValueError("The file type is unsupported")
            else:
                raise ValueError(
                    f"The {doc_type} file type does not match the Loader! Use a suitable one."
                )
        self.byte_content = byte_content
        self._logger = parsing_logger or ParsingLogger(
            silent_errors=silent_errors, name=__name__
        )
        self.pdf_parser = PDFParser(
            pdf_parsing_scheme,
            pdf_extract_images,
            pdf_extract_tables,
            pdf_extract_formulas,
            pdf_remove_service_info,
        )
        self.word_doc_parser = WordDocumentParser(
            word_doc_parsing_scheme,
            word_doc_extract_images,
            word_doc_extract_tables,
            word_doc_extract_formulas,
            timeout_for_converting,
        )
        self._exclude_names = [Path(file).name for file in exclude_files]

    @property
    def logs(self):
        return self._logger.logs

    def lazy_load(
        self,
    ) -> Iterator[Document]:
        """Lazy load given path"""
        content = self.byte_content or self.file_path
        with zipfile.ZipFile(content) as z:
            for info in tqdm(z.infolist(), desc="Zip processing", ncols=80):
                file_name = Path(correct_path_encoding(info.filename))
                if file_name.name in self._exclude_names:
                    continue
                path = str(Path(self.file_path, file_name))
                doc_type = BaseParser.get_doc_type(file_name)
                match doc_type:
                    case DocType.pdf:
                        _parser = self.pdf_parser
                    case DocType.docx | DocType.doc | DocType.odt | DocType.rtf:
                        _parser = self.word_doc_parser
                    case _:
                        if Path(file_name).suffix:
                            self._logger.info(
                                f"Skip file processing in zip, no suitable parser for {path}"
                            )
                        continue

                self._logger.info(f"Processing file in zip: {path}")
                blob = Blob.from_data(
                    z.open(info).read(), path=path, mime_type=doc_type.value
                )
                with self._logger.parsing_info_handler(path):
                    yield from _parser.lazy_parse(blob)



import ast

from langchain_core.language_models import LLM
from langchain_core.prompts import PromptTemplate


class Planner:

    def __init__(self, llm: LLM, prompt_template: PromptTemplate) -> None:
        self._prompt_template = prompt_template
        self._llm = llm

    def generate_answer(self, query: str | list[str]) -> list:
        if isinstance(query, str):
            query = [query]
        inst_query = [self._prompt_template.format(context=i) for i in query]
        answer = [self._llm.invoke(prompt) for prompt in inst_query]
        good_answers, bad_answers = self._extract_planner_queries(answer)
        if len(bad_answers) > 0:
            updated_answers = self._regenerate_answer(bad_answers)
            good_answers += updated_answers
        return good_answers

    def _extract_planner_queries(self, answer: list):
        bad_idx = []
        for i, ans in enumerate(answer):
            try:
                result = ast.literal_eval(ans.content.split('ЗАПРОСЫ:')[1].strip().split(']')[0] + ']')
                answer[i] = result
            except:
                bad_idx.append(i)
        return [answer[i] for i in range(len(answer)) if i not in bad_idx], [answer[i] for i in range(len(answer)) if
                                                                             i in bad_idx]

    def _regenerate_answer(self, query: list, retries: int = 3):
        fixed_queries = []
        for trial in range(retries):
            result = [self._llm.invoke(prompt) for prompt in query]
            good_res, bad_res = self._extract_planner_queries(result)
            fixed_queries += good_res
            query = bad_res
            if len(bad_res) == 0:
                return fixed_queries
        return fixed_queries



import logging
from typing import List, Optional, Any, Dict

import requests
from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    BaseMessage,
    AIMessage,
    HumanMessage,
    SystemMessage,
    ChatResult,
    ChatGeneration,
)
from pydantic import PrivateAttr


# Define the custom LLM class
class Llama31ChatModel(BaseChatModel):
    api_key: str
    base_url: str
    model: str
    temperature: float = 0.5
    max_tokens: int = 3000
    logging_level: int = logging.INFO

    _logger: logging.Logger = PrivateAttr()

    class Config:
        arbitrary_types_allowed = True  # Allows arbitrary types like the logger

    def __init__(self, **data):
        super().__init__(**data)
        self._logger = logging.getLogger(__name__)
        logging.basicConfig(level=self.logging_level)

    @property
    def _llm_type(self) -> str:
        return "llama31"
    
    def _prepare_headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _prepare_context(self, messages: List[BaseMessage]) -> List[Dict[str, str]]:
        role_map = {
            HumanMessage: "user",
            AIMessage: "assistant",
            SystemMessage: "system"
        }
        
        return [{"role": role_map.get(type(message), "user"), "content": message.content} for message in messages]

    def _prepare_payload(
        self,
        context: List[Dict[str, str]],
        stop: Optional[List[str]] = None,
        **kwargs: Any
    ) -> Dict[str, Any]:
        payload = {
            "model": self.model,
            "messages": context,
            "temperature": kwargs.get("temperature", self.temperature),
            "max_tokens": kwargs.get("max_tokens", self.max_tokens),
        }
        if stop is not None:
            payload["stop"] = stop
        return payload

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> ChatResult:
        headers = self._prepare_headers()
        context = self._prepare_context(messages)
        payload = self._prepare_payload(context, stop, **kwargs)

        try:
            response = requests.post(
                f"{self.base_url}/chat/completions",
                headers=headers,
                json=payload,
            )
            response.raise_for_status()
            assistant_response = response.json()["choices"][0]["message"]["content"]
            self._logger.info("REQUEST: %s", payload)
            self._logger.info("RESPONSE: %s", assistant_response)
            ai_message = AIMessage(content=assistant_response)
            generation = ChatGeneration(message=ai_message)
            return ChatResult(generations=[generation])
        except requests.RequestException as e:
            self._logger.error("API request failed: %s", e)
            raise RuntimeError(f"API request failed: {e}")



from pathlib import Path
from typing import Iterator, Union, Any, Optional, Sequence

from langchain_community.document_loaders.directory import _is_visible
from langchain_core.document_loaders import BaseLoader
from langchain_core.documents import Document
from tqdm import tqdm

from protollm.raw_data_processing.docs_parsers.parsers import ParsingScheme, DocType, BaseParser
from protollm.raw_data_processing.docs_parsers.utils.logger import ParsingLogger
from protollm.raw_data_processing.docs_parsers.loaders.pdf_loader import PDFLoader
from protollm.raw_data_processing.docs_parsers.utils.utilities import correct_path_encoding
from protollm.raw_data_processing.docs_parsers.loaders.doc_loader import WordDocumentLoader
from protollm.raw_data_processing.docs_parsers.loaders.zip_loader import ZipLoader


class RecursiveDirectoryLoader(BaseLoader):
    """
    Load files from directory into list of documents.
    """

    def __init__(
        self,
        file_path: Union[str, Path],
        pdf_parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        pdf_extract_images: bool = False,
        pdf_extract_tables: bool = False,
        pdf_extract_formulas: bool = False,
        pdf_remove_service_info: bool = False,
        word_doc_parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        word_doc_extract_images: bool = False,
        word_doc_extract_tables: bool = False,
        word_doc_extract_formulas: bool = False,
        timeout_for_converting: Optional[int] = None,
        exclude_files: Sequence[Union[Path, str]] = (),
        parsing_logger: Optional[ParsingLogger] = None,
        silent_errors: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize with a directory path."""
        self.file_path = Path(file_path)
        if not self.file_path.exists():
            raise FileNotFoundError(f"Directory not found: '{self.file_path}'")
        if not self.file_path.is_dir():
            raise ValueError(f"Expected directory, got file: '{self.file_path}'")

        self._logger = parsing_logger or ParsingLogger(
            silent_errors=silent_errors, name=__name__
        )

        self._pdf_kwargs = {
            "parsing_scheme": pdf_parsing_scheme,
            "extract_images": pdf_extract_images,
            "extract_tables": pdf_extract_tables,
            "extract_formulas": pdf_extract_formulas,
            "remove_headers": pdf_remove_service_info,
            "parsing_logger": self._logger,
        }
        self._word_doc_kwargs = {
            "parsing_scheme": word_doc_parsing_scheme,
            "extract_images": word_doc_extract_images,
            "extract_tables": word_doc_extract_tables,
            "extract_formulas": word_doc_extract_formulas,
            "timeout_for_converting": timeout_for_converting,
            "parsing_logger": self._logger,
        }
        self._zip_kwargs = {
            "pdf_parsing_scheme": pdf_parsing_scheme,
            "pdf_extract_images": pdf_extract_images,
            "pdf_extract_tables": pdf_extract_tables,
            "pdf_extract_formulas": pdf_extract_formulas,
            "pdf_remove_service_info": pdf_remove_service_info,
            "word_doc_parsing_scheme": word_doc_parsing_scheme,
            "word_doc_extract_images": word_doc_extract_images,
            "word_doc_extract_tables": word_doc_extract_tables,
            "word_doc_extract_formulas": word_doc_extract_formulas,
            "timeout_for_converting": timeout_for_converting,
            "exclude_files": exclude_files,
            "parsing_logger": self._logger,
        }
        self._exclude_names = [Path(file).name for file in exclude_files]

    @property
    def logs(self):
        return self._logger.logs

    def lazy_load(
        self,
    ) -> Iterator[Document]:
        """Lazy load given path"""
        paths = [
            path
            for path in self.file_path.rglob("**/[!.]*")
            if path.is_file()
            and _is_visible(path)
            and path.name not in self._exclude_names
            and correct_path_encoding(path.name) not in self._exclude_names
        ]
        for path in tqdm(paths, desc="Directory processing", ncols=80):
            doc_type = BaseParser.get_doc_type(path)
            match doc_type:
                case DocType.pdf:
                    _loader = PDFLoader(path, **self._pdf_kwargs)
                case DocType.docx | DocType.doc | DocType.odt | DocType.rtf:
                    _loader = WordDocumentLoader(path, **self._word_doc_kwargs)
                case DocType.zip:
                    _loader = ZipLoader(path, **self._zip_kwargs)
                case _:
                    self._logger.info(
                        f"Skip file processing, no suitable loader for {path}"
                    )
                    continue

            self._logger.info(f"Processing file: {path}")
            yield from _loader.lazy_load()



from enum import Enum
import uuid
import requests
from websockets import ConnectionClosed
import websockets.sync.client as wsclient
import json
import click
import pandas as pd


class AnswerType(str, Enum):
    RETRIEVAL = 'retrieval'
    ANSWER = 'answer'
    ERROR = 'error'

def parse_ws_response(response):
    response_body = json.loads(response)
    match (name := response_body.get('name')):
        case AnswerType.RETRIEVAL | AnswerType.ANSWER:
            return name, response_body['result']
        case AnswerType.ERROR:
            raise Exception(response_body.get('detail'))
        case _:
            return None, None
        

@click.command()
@click.option("--basepath", type=str, default="0.0.0.0:8080")
@click.option("--output", type=str, default="agents_responses.csv")
def main(basepath, output):
    run_uid = str(uuid.uuid4())
    questions = [
        "Какой объем финансирования программы политики защиты окружающей среды",
        "Какая разница в объеме финансирования программ защиты окружающей среды и образования?",
        "Кто ответственный исполнитель программы политики защиты окружающей среды?",
        "Какие целевые показатели госпрограмм по образованию и защите окружающей среды?",
        "Сколько подпрограмм в госполитике по защите окружающей среды?",
        "Какие приоритеты политики в плане обращений с твердыми коммунальными отходами?",
        "какой объем финансирования программы образования в 2017 году?",
    ]

    response = requests.get(f"http://{basepath}/", params={"agent_type": "streaming"})
    assert response.status_code == 200, "Failed to get agents"
    response = response.json()
    assert len(response) > 0, "No agents found"
    agents_ids = {agent['agent_id']: agent['name'] for agent in response if "rag" in agent['name']}

    agents_responses = list()
    for question in questions:
        question_columns = dict(question=question)
        for agent_id, agent_name in agents_ids.items():
            click.echo(f"Collecting response from {agent_name}, {question=}")
            with wsclient.connect(f"ws://{basepath}/agent") as ws:
                request_payload = {
                    "dialogue_id": run_uid,
                    "agent_id": agent_id,
                    "chat_history":[],
                    "query": question,
                    "run_params": {}
                }
                ws.send(json.dumps(request_payload))
                try:
                    while True:
                        response = ws.recv()
                        response_type, response_data = parse_ws_response(response)
                        if response_type == AnswerType.RETRIEVAL:
                            question_columns[f'docs_{agent_name}'] = response_data
                        elif response_type == AnswerType.ANSWER:
                            question_columns[f'answer_{agent_name}'] = response_data
                except ConnectionClosed:
                    pass
        click.echo("Finished collecting RAG agents responses")

        for endpoint in ('router', 'ensemble'):
            click.echo(f"Collecting response from {endpoint}, {question=}")
            with wsclient.connect(f"ws://{basepath}/{endpoint}") as ws:
                request_payload = {
                    "dialogue_id": run_uid,
                    "chat_history":[],
                    "query": question,
                }
                ws.send(json.dumps(request_payload))
                try:
                    while True:
                        response = ws.recv()
                        response_type, response_data = parse_ws_response(response)
                        if response_type == AnswerType.RETRIEVAL:
                            question_columns[f'docs_{endpoint}'] = response_data
                        elif response_type == AnswerType.ANSWER:
                            question_columns[f'answer_{endpoint}'] = response_data
                except ConnectionClosed:
                    pass
        click.echo("Finished collecting router and ensemble responses")
        click.echo(f"Collected question_columns: {question_columns}")
        agents_responses.append(question_columns)
    click.echo("Finished collecting agents responses")

    df = pd.DataFrame().from_records(agents_responses)
    df.to_csv(output, index=False)

if __name__ == "__main__":
    main()



GEVAL_PROMPT = """criteria=(
    1. Correctness and Relevance:
    - Compare the actual response against the expected response. 
    Determine the extent to which the actual response 
    captures the key elements and concepts of the expected response.
    - Assign higher scores to actual responses that accurately reflect 
    the core information of the expected response, even if only partial.
    2. Numerical Accuracy and Interpretation:
    - Pay particular attention to any  numerical values present 
    in the expected  response. Verify that these values are 
    correctly included in the actual  response and accurately 
    interpreted within the context.
    - Ensure that units of measurement, scales, and numerical 
    relationships are preserved and correctly conveyed.
    3. Allowance for Partial Information:
    - Do not heavily penalize the actual response for incompleteness 
    if it covers significant aspects of the expected response. 
    Prioritize the correctness of provided information over 
    total completeness.
    4. Handling of Extraneous Information:
    - While additional information not present in the expected response 
    should not  necessarily reduce score, 
    ensure that such additions do not introduce inaccuracies 
    or deviate from the context of the expected response.)"""


from pathlib import Path
from typing import Iterator, Union, Any, Optional

from langchain_core.document_loaders import BaseLoader, Blob
from langchain_core.documents import Document

from protollm.raw_data_processing.docs_parsers.parsers import PDFParser, ParsingScheme, DocType
from protollm.raw_data_processing.docs_parsers.utils.logger import ParsingLogger


class PDFLoader(BaseLoader):
    """
    Load PDF into list of documents.
    """

    def __init__(
        self,
        file_path: Union[str, Path],
        byte_content: Optional[bytes] = None,
        parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        extract_images: bool = False,
        extract_tables: bool = False,
        extract_formulas: bool = False,
        remove_headers: bool = False,
        parsing_logger: Optional[ParsingLogger] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize with a file path."""
        self.file_path = str(file_path)
        doc_type = PDFParser.get_doc_type(self.file_path)
        if doc_type is not DocType.pdf:
            if doc_type is DocType.unsupported:
                raise ValueError("The file type is unsupported")
            else:
                raise ValueError(
                    f"The {doc_type} file type does not match the Loader! Use a suitable one."
                )
        self.byte_content = byte_content
        self._logger = parsing_logger or ParsingLogger(name=__name__)
        self.parser = PDFParser(
            parsing_scheme,
            extract_images,
            extract_tables,
            extract_formulas,
            remove_headers,
        )

    @property
    def logs(self):
        return self._logger.logs

    def lazy_load(
        self,
    ) -> Iterator[Document]:
        """Lazy load given path"""
        if self.byte_content is None:
            blob = Blob.from_path(self.file_path)
        else:
            blob = Blob.from_data(
                self.byte_content, path=self.file_path, mime_type=DocType.pdf.value
            )
        with self._logger.parsing_info_handler(self.file_path):
            yield from self.parser.lazy_parse(blob)



from pathlib import Path
from typing import Union

import chardet
from ftfy import is_bad


def is_bad_encoding(lines: list[str]) -> bool:
    count_bad = sum([_is_bad(line) for line in lines], start=0)
    proportion_bad = count_bad / max(1, len(lines))
    return proportion_bad >= 0.5


def _is_bad(text: str) -> bool:
    if is_bad(text):
        return True
    # This verification works for russian docs !!!
    try:
        text.encode("sloppy-windows-1252")
    except UnicodeEncodeError:
        return False
    return True


def correct_path_encoding(path: Union[str, Path]) -> str:
    path = Path(path)
    path = Path(*[fix_zip_path(part) for part in path.parts])
    return str(path)


def fix_zip_path(path: str) -> str:
    try:
        string_bytes = path.encode("437")
        guessed_encoding = chardet.detect(string_bytes)["encoding"] or "cp1252"
        path = string_bytes.decode(guessed_encoding, "replace")
    except:
        pass
    return path



import re
from typing import Any, Iterable, Optional

from langchain_text_splitters import TextSplitter

from protollm.raw_data_processing.docs_transformers.utilities import fix_list_dots_separators


class SentencesSplitter(TextSplitter):
    def __init__(self, separators: Optional[Iterable[str]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._separators = separators or [r"\. (?<![0-9]\. | ^[0-9].)"]

    def split_text(self, text: str) -> list[str]:
        sentences_lst = [
            x.strip() for x in re.split(" | ".join(self._separators), text)
        ]
        return fix_list_dots_separators(sentences_lst)



from typing import Any, List

from pydantic import BaseModel


class ConfigLoader(BaseModel):
    file_path: str = ''
    save_path: str = ''
    loader_name: str
    parsing_params: dict[str, Any] = dict()


class ConfigSplitter(BaseModel):
    splitter_name: str | None = None
    splitter_params: dict[str, Any] = dict()


class ConfigFile(BaseModel):
    loader: ConfigLoader
    splitter: List[ConfigSplitter] = []
    tokenizer: str | None = None



from protollm.rags.stores.elasticsearch.utilities import (get_index_name, get_elasticsearch_store,
                                                       custom_query_for_metadata_mapping)
from protollm.rags.stores.elasticsearch.retrieval_strategies import BM25RetrievalStrategy



import re

from protollm.raw_data_processing.docs_parsers.parsers.utilities import (
    HEADING_KEYWORDS,
    HEADING_STOP_LIST,
)


def _get_heading_hierarchy_level(text: str, metadata: dict) -> int:
    low_text = text.lower()
    if any([stop_word in low_text for stop_word in HEADING_STOP_LIST]):
        return -1

    low_text = "".join(re.findall(r"[^\W\d_]", low_text, re.UNICODE))
    if low_text in HEADING_KEYWORDS:
        return 0

    if metadata["bold"]:
        if metadata["list_level"] != -1:
            return metadata["list_level"]
        if metadata["is_bullet_list"] or text.isupper():
            return 0

    return -1


def add_headings_hierarchy(
    lines: list[str], metadata: list[dict]
) -> tuple[list[str], list[dict]]:
    new_lines, new_metadata = [], []
    hierarchy = [""]
    for line, line_meta in zip(lines, metadata):
        hierarchy_level = _get_heading_hierarchy_level(line, line_meta)
        if hierarchy_level == -1:
            new_meta = {**line_meta, "headings": list(hierarchy)}
            new_lines.append(line)
            new_metadata.append(new_meta)
        else:
            if hierarchy_level < len(hierarchy):
                hierarchy = hierarchy[:hierarchy_level]
            elif hierarchy_level > len(hierarchy):
                hierarchy.extend([""] * (hierarchy_level - len(hierarchy)))
            hierarchy.append(line)
    return new_lines, new_metadata


def get_chapters(
    lines: list[str], metadata: list[dict]
) -> tuple[list[str], list[dict]]:
    chapters, meta = [], []
    cur_chapter = ""
    for line, line_meta in zip(lines, metadata):
        if chapters and (cur_chapter == line_meta["headings"][0]):
            chapters[-1] = join_texts(chapters[-1], line, joiner="\n")
            update_metadata(meta[-1], line_meta)
            continue
        cur_chapter = line_meta["headings"][0]
        chapters.append(line)
        meta.append({"chapter": cur_chapter, "headings": [cur_chapter]})
        update_metadata(meta[-1], line_meta)
    return chapters, meta


def get_paragraphs(
    lines: list[str], metadata: list[dict]
) -> tuple[list[str], list[dict]]:
    paragraphs, meta = [], []
    cur_paragraph = ""
    for line, line_meta in zip(lines, metadata):
        if (
            paragraphs
            and line_meta["is_bullet_list"]
            and (cur_paragraph == line_meta["headings"][-1])
        ):
            paragraphs[-1] = join_texts(paragraphs[-1], line, joiner=" ")
            update_metadata(meta[-1], line_meta)
            continue
        cur_paragraph = line_meta["headings"][-1]
        paragraphs.append(line)
        meta.append({"headings": line_meta["headings"]})
        update_metadata(meta[-1], line_meta)
    return paragraphs, meta


def update_metadata(meta: dict, line_meta: dict):
    for key in ["images", "formulas", "urls", "tables"]:
        meta[key] = {**meta.get(key, {}), **line_meta.get(key, {})}


def join_texts(original: str, new: str, joiner: str = "\n") -> str:
    return joiner.join((original, new))



DOMAIN_SPECIFIC_ASSTISTANT="""
You are a smart AI assistant. You have high expertise in the field 
    [...]. 
    Answer the question following the rules below.
    1. Before giving an answer to the user question, provide an 
    explanation. Mark the answer with keyword 'ANSWER', and 
    explanation with 'EXPLANATION'. Both answer and explanation must be 
    in the English language.
    2. If the question is about complaints, answer about at least 5 
    complaints topics.
    3. Answer should be five sentences maximum.
    4. In answers you must use only the English language.
"""




from typing import Optional, Any

from langchain_elasticsearch import ElasticsearchStore
from langchain_elasticsearch.vectorstores import BaseRetrievalStrategy

from protollm.rags.stores.elasticsearch.settings import settings
from protollm.rags.stores.elasticsearch.retrieval_strategies import BM25RetrievalStrategy


def get_index_name(index: int) -> str:
    return f"index_v{index}"


def get_elasticsearch_store(index_name: str,
                            es_url: str = settings.es_url,
                            es_user: str = settings.es_user,
                            es_password: str = settings.es_password,
                            query_field: str = settings.content_field,
                            strategy: BaseRetrievalStrategy = BM25RetrievalStrategy(),
                            es_params: Optional[dict[str, Any]] = None) -> ElasticsearchStore:
    return ElasticsearchStore(index_name,
                              es_url=es_url,
                              es_user=es_user,
                              es_password=es_password,
                              query_field=query_field,
                              strategy=strategy,
                              es_params=es_params)


def custom_query_for_metadata_mapping(query_body: dict, query: str) -> dict:
    """Custom query to be used in Elasticsearch with indexes that use langchain Document schema.
    This implies that all additional fields are stored in the metadata
    Args:
        query_body (dict): Elasticsearch query body.
        query (str): Query string.
    Returns:
        dict: Elasticsearch query body.
    """
    query = query_body['query']
    if 'multi_match' in query:
        if 'fields' in query['multi_match']:
            fields = []
            for field in query['multi_match']['fields']:
                if not (field.startswith('metadata') or field.startswith(settings.content_field)):
                    field = 'metadata.' + field
                fields.append(field)

            query['multi_match']['fields'] = fields

    return query_body



import json
from os.path import dirname
from pathlib import Path

from pydantic import computed_field
from pydantic_settings import BaseSettings, SettingsConfigDict


CONFIG_PATH = Path(dirname(dirname(__file__)), '/stores/elasticsearch/configs')


class ElasticsearchSettings(BaseSettings):
    es_host: str = "elasticDUMB"
    es_port: int = 9201

    es_user: str = "elastic"
    es_password: str = "admin"

    @computed_field
    @property
    def es_url(self) -> str:
        return f"http://{self.es_host}:{self.es_port}"

    es_index_mappings: dict = json.loads(Path(CONFIG_PATH, 'index_mappings.json').read_text(encoding="utf-8"))
    es_index_settings: dict = json.loads(Path(CONFIG_PATH, 'index_settings.json').read_text(encoding="utf-8"))
    es_query_template: dict = json.loads(Path(CONFIG_PATH, 'query_template.json').read_text(encoding="utf-8"))
    es_query_all_hits: dict = json.loads(Path(CONFIG_PATH, 'query_all_hits.json').read_text(encoding="utf-8"))

    metadata_fields: list[str] = list(es_index_mappings['properties']['metadata']['properties'].keys())
    content_field: str = 'paragraph'

    model_config = SettingsConfigDict(
        env_file=Path(Path(__file__).parent.parent, '/configs/elastic.env'),
        env_file_encoding='utf-8',
        extra='ignore',
    )


settings = ElasticsearchSettings()



import os
from pathlib import Path
from typing import Any

from protollm.raw_data_processing.docs_parsers.loaders import PDFLoader, WordDocumentLoader, ZipLoader, \
    RecursiveDirectoryLoader
from langchain_core.document_loaders import BaseLoader

from protollm.rags.pipeline.docs_processing.entities import LoaderType, LangChainDocumentLoader
from protollm.rags.pipeline.docs_processing.exceptions import FileExtensionError, PathIsNotAssigned


def get_loader(**loader_params) -> BaseLoader:
    document_path = loader_params.get('file_path', '')
    if not isinstance(document_path, (str, Path)) or document_path == '':
        raise PathIsNotAssigned('Input file (directory) path is not assigned')
    doc_extension = str(document_path).lower().split('.')[-1]
    if os.path.isdir(document_path):
        doc_extension = LoaderType.directory
    match doc_extension:
        case LoaderType.pdf:
            return PDFLoader(**loader_params)
        case LoaderType.json:
            return LangChainDocumentLoader(**loader_params)
        case LoaderType.docx | LoaderType.doc | LoaderType.rtf | LoaderType.odt:
            return WordDocumentLoader(**loader_params)

    parsing_scheme = loader_params.pop('parsing_scheme', 'lines')
    extract_images = loader_params.pop('extract_images', False)
    extract_tables = loader_params.pop('extract_tables', False)
    parse_formulas = loader_params.pop('parse_formulas', False)
    remove_service_info = loader_params.pop('remove_service_info', False)
    loader_params = dict(
        pdf_parsing_scheme=parsing_scheme,
        pdf_extract_images=extract_images,
        pdf_extract_tables=extract_tables,
        pdf_parse_formulas=parse_formulas,
        pdf_remove_service_info=remove_service_info,
        word_doc_parsing_scheme=parsing_scheme,
        word_doc_extract_images=extract_images,
        word_doc_extract_tables=extract_tables,
        word_doc_parse_formulas=parse_formulas,
        word_doc_remove_service_info=remove_service_info,
        **loader_params,
    )
    match doc_extension:
        case LoaderType.zip:
            return ZipLoader(**loader_params)
        case LoaderType.directory:
            return RecursiveDirectoryLoader(**loader_params)
        case _:
            raise FileExtensionError(f'File with extension {doc_extension} has not been implemented yet.')



class PathIsNotAssigned(Exception):
    def __init__(self, message):
        super.__init__(message)


class PipelineError(Exception):
    def __init__(self, message):
        super().__init__(message)


class FileExtensionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class TransformerNameError(Exception):
    def __init__(self, message):
        super().__init__(message)


class LoaderNameError(Exception):
    def __init__(self, message):
        super().__init__(message)


import mimetypes
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterator, List, Union

from langchain_core.document_loaders import Blob
from langchain_core.documents import Document

from protollm.raw_data_processing.docs_parsers.parsers.entities import DocType


class BaseParser(ABC):
    """
    Abstract interface for parsers.

    A parser provides a way to parse raw data into one or more documents.
    """

    @abstractmethod
    def lazy_parse(self, blob: Blob) -> Iterator[Document]:
        """Lazy parsing interface.

        Subclasses are required to implement this method.

        Args:
            blob: representation of raw data from file

        Returns:
            Generator of documents
        """

    def parse(self, blob: Blob) -> List[Document]:
        """Eagerly parse raw data into a document or documents.

        This is a convenience method for interactive development environment.

        Production applications should favor the lazy_parse method instead.

        Subclasses should generally not over-ride this parse method.

        Args:
            blob: representation of raw data from file

        Returns:
            List of documents
        """
        return list(self.lazy_parse(blob))

    @staticmethod
    def get_doc_type(file: Union[str, Path]) -> DocType:
        mimetype = mimetypes.guess_type(file)[0]
        if mimetype is None:
            mimetype = Path(file).suffix.replace(".", "")

        match mimetype:
            case "application/pdf" | "pdf":
                return DocType.pdf
            case (
                "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
                | "docx"
            ):
                return DocType.docx
            case "application/msword" | "doc":
                return DocType.doc
            case "application/vnd.oasis.opendocument.text" | "odt":
                return DocType.odt
            case "application/rtf" | "rtf":
                return DocType.rtf
            case (
                "application/zip"
                | "application/x-zip-compressed"
                | "multipart/x-zip"
                | "zip"
            ):
                return DocType.zip
            case _:  # TODO: add txt, xlsx, pptx support
                return DocType.unsupported



from enum import Enum


_namespace_mapping = {
    "w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
    "m": "http://schemas.openxmlformats.org/officeDocument/2006/math",
    "a": "http://schemas.openxmlformats.org/drawingml/2006/main",
    "r": "http://schemas.openxmlformats.org/officeDocument/2006/relationships",
}


def _get_xml_tag_name(tag: str, namespace: str) -> str:
    namespace = _namespace_mapping[namespace]
    return f"{{{namespace}}}{tag}"


class XMLTag(str, Enum):
    raw = _get_xml_tag_name("r", "w")
    text = _get_xml_tag_name("t", "w")
    image = _get_xml_tag_name("drawing", "w")
    blip = _get_xml_tag_name("blip", "a")
    embed = _get_xml_tag_name("embed", "r")
    math = _get_xml_tag_name("oMath", "m")
    math_paragraph = _get_xml_tag_name("oMathPara", "m")



import json
from typing import Any, Dict, List, Optional, Union

import requests
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (AIMessage, BaseMessage, HumanMessage,
                                     SystemMessage)
from langchain_core.outputs import ChatGeneration, ChatResult


class ChatRESTServer(BaseChatModel):
    model: Optional[str] = 'llama3'
    base_url: str = 'http://10.32.2.2:8672'
    timeout: Optional[int] = None
    """Timeout for the request stream"""

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "chat-rest-server"

    def _convert_messages_to_rest_server_messages(
            self, messages: List[BaseMessage]
    ) -> List[Dict[str, Union[str, List[str]]]]:
        chat_messages: List = []
        for message in messages:
            role = ""
        match message:
            case HumanMessage():
                role = "user"
            case AIMessage():
                role = "assistant"
            case SystemMessage():
                role = "system"
            case _:
                raise ValueError("Received unsupported message type.")


        content = ""
        if isinstance(message.content, str):
            content = message.content
        else:
            raise ValueError(
                "Unsupported message content type. "
                "Must have type 'text' "
            )
        chat_messages.append(
            {
                "role": role,
                "content": content
            }
        )
        return chat_messages

    def create_chat(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> Dict[str, Any]:
        payload = {
            "model": self.model,
            "messages": self._convert_messages_to_rest_server_messages(
                messages)
        }

        if self.base_url=='mock':
            return {'test':'test'}

        response = requests.post(
            url=f'{self.base_url}/v1/chat/completions',
            headers={"Content-Type": "application/json"},
            json=payload,
            timeout=self.timeout
        )
        response.encoding = "utf-8"
        match response.status_code:
            case 200:
                pass  # Status code is 200, no action needed
            case 404:
                raise ValueError(
                    "CustomWeb call failed with status code 404. "
                    "Maybe you need to connect to the corporate network."
                )
            case _:
                optional_detail = response.text
                raise ValueError(
                    f"CustomWeb call failed with status code "
                    f"{response.status_code}. "
                    f"Details: {optional_detail}"
                )
        return json.loads(response.text)

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> ChatResult:
        response = self._create_chat(messages, stop, **kwargs)
        chat_generation = ChatGeneration(
            message=AIMessage(
                content=response['choices'][0]['message']['content']),
            generation_info=response,
        )
        return ChatResult(generations=[chat_generation])

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters.

        This information is used by the LangChain callback system, which
        is used for tracing purposes make it possible to monitor LLMs.
        """
        return {
            "model_name": self.model,
            "url": self.base_url
        }



import re
import warnings
from collections import Counter
from functools import reduce

import PyPDF2
import numpy as np
import pdfplumber
import pytesseract
from PIL import Image
from ftfy import fix_text
from pdf2image import convert_from_path
from pdfminer.high_level import extract_pages
from pdfminer.layout import LTTextContainer, LTChar, LTFigure, LTTextLine, LAParams
from tabulate import tabulate

from protollm.raw_data_processing.docs_parsers.utils.exceptions import (
    NoTextLayerError,
    ParseImageWarning,
    TitleExtractingWarning,
    PageNumbersExtractingWarning,
)
from protollm.raw_data_processing.docs_parsers.parsers.utilities import (
    HEADING_KEYWORDS,
    HEADING_STOP_LIST,
    FOOTER_KEYWORDS,
)

listmerge = lambda s: reduce(lambda d, el: d.extend(el) or d, s, [])


def text_extraction(element):
    """
    Main text extracting function from all elements in document layout
    :param element: element from the document's layout
    :return:
    """
    line_text = element.get_text()

    line_formats = []
    for text_line in element:
        if isinstance(text_line, LTTextContainer):
            for character in text_line:
                if isinstance(character, LTChar):
                    line_formats.append(character.fontname)
                    line_formats.append(character.size)
    format_per_line = list(set(line_formats))
    return line_text, format_per_line


def extract_table(stream, page_num, table_num):
    """
    Function for extracting tables from the page
    :param stream: binary input
    :param page_num:
    :param table_num:
    :return:
    """
    pdf = pdfplumber.open(stream)
    table_page = pdf.pages[page_num]
    table = table_page.extract_tables()[table_num]

    return table


def convert_table_to_html(table) -> str:
    """
    Converts table to string in html format
    :param table:
    :return:
    """
    processed_table = [
        [item.replace("\n", " ") if item is not None else "" for item in row]
        for row in table
    ]

    return tabulate(processed_table, headers="firstrow", tablefmt="html")


def is_element_inside_any_table(element, page, tables):
    """
    Checks if the element is in any tables present in the page
    :param element:
    :param page:
    :param tables:
    :return:
    """
    x0, y0up, x1, y1up = element.bbox

    y0 = page.bbox[3] - y1up
    y1 = page.bbox[3] - y0up
    for table in tables:
        tx0, ty0, tx1, ty1 = table.bbox
        if tx0 <= x0 <= x1 <= tx1 and ty0 <= y0 <= y1 <= ty1:
            return True
    return False


def find_table_for_element(element, page, tables):
    """
    Find the table's index for a given element
    :param element:
    :param page:
    :param tables:
    :return:
    """
    x0, y0up, x1, y1up = element.bbox
    y0 = page.bbox[3] - y1up
    y1 = page.bbox[3] - y0up
    for i, table in enumerate(tables):
        tx0, ty0, tx1, ty1 = table.bbox
        if tx0 <= x0 <= x1 <= tx1 and ty0 <= y0 <= y1 <= ty1:
            return i
    return None


def crop_image(element, pageObj):
    """
    Crops the image elements from PDFs
    :param element:
    :param pageObj:
    :return:
    """
    [image_left, image_top, image_right, image_bottom] = [
        element.x0,
        element.y0,
        element.x1,
        element.y1,
    ]

    pageObj.mediabox.lower_left = (image_left, image_bottom)
    pageObj.mediabox.upper_right = (image_right, image_top)

    cropped_pdf_writer = PyPDF2.PdfWriter()
    cropped_pdf_writer.add_page(pageObj)

    with open("cropped_image.pdf", "wb") as cropped_pdf_file:
        cropped_pdf_writer.write(cropped_pdf_file)


def convert_to_images(
    input_file,
):
    """
    Converts the PDF to images
    :param input_file:
    :return:
    """
    images = convert_from_path(input_file)
    image = images[0]
    output_file = "PDF_image.png"
    image.save(output_file, "PNG")


def image_to_text(image_path):
    # Read the image
    img = Image.open(image_path)
    # Extract the text from the image
    text = pytesseract.image_to_string(img)
    return text


def get_heading_info(element_info, heading_env, doc_info):
    element_text = element_info["element"].get_text().replace("\n", " ").strip()
    for stop_word in HEADING_STOP_LIST:
        if stop_word in element_text.lower():
            return -1

    special_symbols_cnt = 0
    letters_cnt = 0
    for char in element_text:
        if char.isalpha():
            letters_cnt += 1
        if not char.isalpha() and not char.isdigit() and not char == " ":
            special_symbols_cnt += 1
    if special_symbols_cnt > 4 or letters_cnt == 0:
        return -1

    if element_text == "":
        return -1

    # Get meta info about the element
    numeric_pref = get_numeric_prefix_str(element_info["element"])
    is_bold = element_info["meta"]["format"]["font_style"] == "bold"
    is_upper = element_info["element"].get_text().isupper()
    line_font = element_info["meta"]["format"]["font_name"]

    if (
        is_bold
    ):  # if the line starts with the numeric prefix and is bold - it is heading of the particular level
        if numeric_pref is not None:
            heading_lvl = len([x for x in numeric_pref.split(".") if x != ""])
            return heading_lvl
        else:
            if heading_env == 1:
                return heading_env  # it is a continuation of the heading
            if is_upper:
                return 1  # it is a first-level heading

    else:  # it is not bold
        if is_upper:
            if numeric_pref is not None:
                heading_lvl = len([x for x in numeric_pref.split(".") if x != ""])
                return heading_lvl
        if line_font != doc_info["font_name"]:
            if numeric_pref is not None:
                heading_lvl = len([x for x in numeric_pref.split(".") if x != ""])
                return heading_lvl
            # elif is_upper:
            #     return 1  # it is a first-level heading
            else:
                if heading_env == 1:
                    return heading_env  # it is a continuation of the heading

    for keyword in HEADING_KEYWORDS:
        element_text = element_info["element"].get_text().lower().strip()
        if element_text == keyword:
            return 1  # it is a first-level heading

    # If the element's font size refers to heading on some level
    element_size = element_info["meta"]["format"]["fontsize"]
    if element_size in doc_info["headings_sizes"]:
        return doc_info["headings_sizes"][element_size]

    return -1  # it is not a heading or a part of the heading


def get_numeric_prefix_str(text_line):
    line_text = text_line.get_text()
    numeric_pref = re.match(
        "^([0-9]+(\.)?)+", line_text
    )  # check if the line starts with the numeric prefix
    if numeric_pref is not None:
        return numeric_pref.group()
    else:
        return None


def check_layout(pages_layout):
    text_lines_stat = []
    for page_number, page in enumerate(pages_layout):
        if page_number > 10:
            break
        # Get text length statistics in all text elements
        for element in page:
            if isinstance(element, LTTextContainer):
                if not isinstance(
                    element, LTTextLine
                ):  # text element is a Box and should be unpacked to Lines
                    for line in element:
                        text_lines_stat.append(
                            len(line.get_text().split(" "))
                        )  # calculate number of words in parsed line
                else:
                    text_lines_stat.append(
                        len(element.get_text().split(" "))
                    )  # calculate number of words in parsed line
    avg_words_number = np.mean(text_lines_stat) if text_lines_stat else 0
    if avg_words_number < 4:
        return -1
    elif avg_words_number > 10:
        return 1
    else:
        return 0


def get_document_layout(stream, layout_parsing_params, tables_by_pages):
    # Get all line elements from the document, grouped by pages
    pages_layout = extract_pages(stream, laparams=layout_parsing_params)
    check_layout_res = check_layout(pages_layout)

    if check_layout_res == 1:
        layout_parsing_params.word_margin = layout_parsing_params.word_margin + 2
    elif check_layout_res == -1:
        layout_parsing_params.word_margin = 0.1
    pages_layout = extract_pages(stream, laparams=layout_parsing_params)

    pages_structure = []  # variable for the updated document's structure by pages

    for page_number, page in enumerate(pages_layout):
        page_structure = []

        # Get all tables from the page
        page_tables = tables_by_pages[page_number]

        # Analyze all elements on the page
        for element in page:
            if isinstance(element, LTTextContainer):
                text_lines_lst = []
                if not isinstance(
                    element, LTTextLine
                ):  # text element is a Box and should be unpacked to Lines
                    for line in element:
                        text_lines_lst.append(line)
                else:
                    text_lines_lst.append(element)

                for line in text_lines_lst:
                    if page_tables is not None:  # if there are any tables on the page
                        if is_element_inside_any_table(line, page, page_tables):
                            table_id_found = find_table_for_element(
                                line, page, page_tables
                            )
                            if (
                                table_id_found is not None
                            ):  # text element is a part of the table
                                page_structure.append(
                                    {
                                        "element": line,
                                        "meta": {"type": "table", "id": table_id_found},
                                    }
                                )
                                continue
                    # text line is not a part of the table
                    page_structure.append(
                        {"element": line, "meta": {"type": "text", "id": -1}}
                    )
            elif isinstance(element, LTFigure):  # element is an image
                page_structure.append(
                    {"element": element, "meta": {"type": "image", "id": -1}}
                )
        pages_structure.append(page_structure)

    # check if the first page is a title page
    # for element_info in pages_structure[0]:
    #     element = element_info['element']
    #     if isinstance(element, LTTextContainer):
    #         for keyword in HEADING_STOP_LIST:
    #             if keyword in element.get_text():
    #                 return pages_structure[1:]  # return document without title
    return pages_structure


def get_document_formatting(pages_structure):
    # margin_inf = 1000  # maximum value of the left margin attribute

    # Initialize main variables for the document formatting and layout structure
    doc_info = {}
    doc_structure = []

    # Variables for the whole document's formatting analysis
    doc_font_size_counter = Counter()
    doc_line_spacing_counter = Counter()
    doc_left_margin_counter = Counter()
    # for the cases, when information about the particular font is unavailable and there are only CID Fonts
    doc_main_font_counter = Counter()

    doc_headings_sizes = []  # list of headings' font sizes in hierarchical order

    for page in pages_structure:
        page_structure = []
        prev_line_bottom_border = (
            None  # y coordinate of the previous line's bottom border
        )

        for element_info in page:
            element_format_info = {}  # dict with info about text formatting
            elem_font_size_counter = Counter()
            elem_font_name_counter = Counter()

            elem_start_symbol = ""
            elem_font_style = "plain"

            # process only text elements (including ones which are parts of tables)
            if (
                element_info["meta"]["type"] != "image"
                and element_info["element"].get_text() != ""
            ):
                element = element_info["element"]
                if prev_line_bottom_border is not None:
                    elem_line_spacing = (
                        prev_line_bottom_border - element.y0
                    )  # space between the lines
                else:
                    elem_line_spacing = -1  # first line on the page

                prev_line_bottom_border = (
                    element.y1
                )  # update previous line bottom border attribute

                elem_left_margin = element.x0  # left margin of the line

                is_bold = True
                no_letters = True

                # Analyze element characters' formatting
                for character in element:
                    if character.get_text()[0].isalpha():
                        no_letters = False
                        if "Bold" not in character.fontname:
                            is_bold = False
                    if elem_start_symbol == "":
                        elem_start_symbol = (
                            "letter"
                            if character.get_text()[0].isalpha()
                            else (
                                "digit"
                                if character.get_text()[0].isdigit()
                                else "symbol"
                            )
                        )
                    if isinstance(character, LTChar):
                        elem_font_size_counter[round(character.size)] += 1  # font size
                        elem_font_name_counter[character.fontname] += 1  # font name

                # Get info about element's main font style (plain/bold)
                # if 'Bold' in elem_font_name_counter.most_common(1)[0][0] and len(elem_font_name_counter) == 1:
                #     elem_font_style = 'bold'
                if is_bold and not no_letters:
                    elem_font_style = "bold"

                # Get info about element's main font size
                elem_font_size = elem_font_size_counter.most_common(1)[0][0]
                elem_font_name = elem_font_name_counter.most_common(1)[0][0]

                # Update info about the element's formatting
                element_format_info["fontsize"] = elem_font_size
                element_format_info["font_style"] = elem_font_style
                element_format_info["font_name"] = elem_font_name
                element_format_info["left_margin"] = elem_left_margin
                element_format_info["line_spacing"] = elem_line_spacing
                element_format_info["start_symbol"] = elem_start_symbol

                # Update document's statistics
                doc_line_spacing_counter[
                    elem_line_spacing
                ] += 1  # add line spacing attribute
                doc_left_margin_counter[
                    elem_left_margin
                ] += 1  # add left margin attribute
                doc_font_size_counter[elem_font_size] += 1  # add font size attribute
                doc_main_font_counter[
                    elem_font_name
                ] += 1  # add font name info attribute

            element_info["meta"][
                "format"
            ] = element_format_info  # set format info for the element
            page_structure.append(element_info)
        doc_structure.append(page_structure)

    doc_font_size = doc_font_size_counter.most_common(1)[0][
        0
    ]  # main document's font size
    doc_line_spacing = doc_line_spacing_counter.most_common(1)[0][
        0
    ]  # main document's line spacing
    doc_left_margins = [
        x[0] for x in doc_left_margin_counter.most_common(2)
    ]  # two most common left margins
    doc_main_font = doc_main_font_counter.most_common(1)[0][0]

    for font_size in [x[0] for x in doc_font_size_counter.items()]:
        if font_size > doc_font_size:
            doc_headings_sizes.append(font_size)
    doc_headings_sizes.sort(reverse=True)  # sort font sizes in the descending order

    # Create dictionary to get heading level by heading font size
    doc_headings_sizes_dict = {}
    for i in range(len(doc_headings_sizes)):
        doc_headings_sizes_dict[doc_headings_sizes[i]] = i + 1

    # Prepare info about whole document's formatting
    doc_info["font_size"] = doc_font_size
    doc_info["font_name"] = doc_main_font
    doc_info["line_spacing"] = doc_line_spacing
    doc_info["left_margin"] = doc_left_margins
    doc_info["headings_sizes"] = doc_headings_sizes_dict

    return doc_info, doc_structure


def is_heading_correct(heading_str):
    if heading_str[0].isalpha() and heading_str[0].islower():
        return False

    if not heading_str[0].isalpha() and not heading_str[0].isdigit():
        return False

    if len(heading_str) > 128:
        return False

    return True


def extract_by_lines(
    stream,
    parse_images=False,
    parse_tables=True,
    parse_formulas=False,
    remove_service_info=False,
) -> tuple[list[str], list[dict]]:
    """
    Parses given pdf document to lines content and meta
    :param parse_images:
    :param parse_tables:
    :param parse_formulas:
    :param stream:
    :return:
    """
    document_content = []
    document_meta = []

    # Set up hyperparameters for the document's layout parsing by lines
    params = LAParams(
        line_overlap=0.5,
        char_margin=15.0,
        line_margin=0.1,
        word_margin=2.0,
        boxes_flow=1.0,
        detect_vertical=False,
        all_texts=True,
    )

    # Tables processing
    tables_by_pages = []
    tables_reader = pdfplumber.open(stream)

    # Extract all tables from the document using pdfplumber
    for page in tables_reader.pages:
        tables_by_pages.append(page.find_tables())

    # Get all line elements, grouped by pages, with the meta about the types: text, table or image
    pages_layout = get_document_layout(stream, params, tables_by_pages)

    is_text_in_doc = False

    for page_layout in pages_layout:
        for element_info in page_layout:
            if (
                element_info["meta"]["type"] == "text"
                or element_info["meta"]["type"] == "table"
            ):
                is_text_in_doc = True
                break
        if is_text_in_doc:
            break

    if not is_text_in_doc:
        raise NoTextLayerError("Document contains no text layer, only images")

    # Set up pdf reader for working with images
    img_pdf_reader = PyPDF2.PdfReader(stream)

    # Get info about the general document's formatting and each element's formatting
    doc_info, doc_structure = get_document_formatting(pages_layout)

    # Set up environmental variables
    heading_env = (
        -1
    )  # level of the current heading's environment (-1 if not the heading's environment)
    heading_lst = []
    current_heading_lvl = -1  # means that the element is not the heading (basic)
    tables_analysed = [set() for _ in range(len(doc_structure))]
    paragraph_id = 0
    headings_hierarchy = []

    for page_number, page in enumerate(doc_structure):
        document_content.append([])
        document_meta.append([])
        if parse_images:
            try:
                page_object = img_pdf_reader.pages[page_number]
            except IndexError:
                warnings.warn(
                    "Error in pages structure formatting", category=ParseImageWarning
                )
                page_object = None

        for element_info in page:
            is_heading = 0
            correct_heading = True
            headings = []
            paragraph = -1

            element = element_info["element"]
            element_meta = element_info["meta"]
            element_text = ""

            if element_meta["type"] == "text":
                element_text = fix_text(element.get_text().replace("\n", " ").strip())
                heading_lvl = get_heading_info(element_info, heading_env, doc_info)

                if heading_lvl != -1:  # text line is a part of some heading
                    if heading_lvl == heading_env:  # continuation
                        heading_lst.append(element_text)  # continue the heading
                    else:  # heading of another level or the new heading after other elements
                        if len(heading_lst) != 0:
                            # we should add info about the previous heading to the 'headings' list
                            headings_hierarchy = headings_hierarchy[: heading_env - 1]
                            heading_str = " ".join(heading_lst).strip()
                            if is_heading_correct(heading_str):
                                headings_hierarchy.append(heading_str.strip())
                            else:
                                correct_heading = False
                        heading_lst = [element_text]  # starting new heading
                        heading_env = heading_lvl

                    is_heading = 1
                else:  # text line is not a part of any heading
                    is_heading = 0
                    if heading_env != -1:  # previous element contained heading
                        paragraph_id += 1  # new paragraph
                        if len(heading_lst) != 0:
                            # we should add info about the previous heading to the 'headings' list
                            headings_hierarchy = headings_hierarchy[: heading_env - 1]
                            heading_str = " ".join(heading_lst).strip()
                            if is_heading_correct(heading_str):
                                headings_hierarchy.append(heading_str)
                            else:
                                correct_heading = False
                            heading_lst = []
                            heading_env = -1
                    else:  # plain text and previous text line was not a heading
                        # figuring out, should we start a new paragraph
                        element_spacing = element_info["meta"]["format"]["line_spacing"]
                        element_margin = element_info["meta"]["format"]["left_margin"]
                        start_symbol = element_info["meta"]["format"]["start_symbol"]

                        # if the line spacing is bigger than average or left margin is bigger than average,
                        # and it is not a list element (starts with letter or digit)
                        if start_symbol != "symbol":
                            if (
                                element_spacing > 1.1 * doc_info["line_spacing"]
                                or element_margin > doc_info["left_margin"][0]
                            ):
                                paragraph_id += 1  # new paragraph

                    headings = headings_hierarchy
                    paragraph = paragraph_id

            else:
                if heading_env != -1:  # previous element contained heading
                    paragraph_id += 1  # new paragraph
                    if len(heading_lst) != 0:
                        # we should add info about the previous heading to the 'headings' list
                        headings_hierarchy = headings_hierarchy[: heading_env - 1]
                        heading_str = " ".join(heading_lst).strip()
                        if is_heading_correct(heading_str):
                            headings_hierarchy.append(heading_str)
                        else:
                            correct_heading = False
                        heading_lst = []
                        heading_env = -1

                if element_meta["type"] == "table":
                    table_id = element_meta["id"]
                    if table_id not in tables_analysed[page_number] and parse_tables:
                        tables_analysed[page_number].add(table_id)
                        table = extract_table(stream, page_number, table_id)

                        # add string with table content to the document content
                        element_text = convert_table_to_html(table)

                        # prepare meta info for the table element
                        is_heading = 0
                        headings = headings_hierarchy
                        paragraph = paragraph_id
                else:  # element_meta['type'] == 'image'
                    if parse_images:
                        if page_object is not None:
                            crop_image(element["element"], page_object)
                            convert_to_images("cropped_image.pdf")
                            image_text = image_to_text("PDF_image.png")

                            # add recognized text from image to the document content
                            element_text = image_text

                        # prepare meta info for the image element
                        is_heading = 0  # only text element can be a heading
                        headings = headings_hierarchy
                        # paragraph_id += 1  # image related to the new paragraph
                        paragraph = paragraph_id

            if element_text != "" and (is_heading != 0 or paragraph != -1):
                if remove_service_info:
                    is_footer = False
                    for footer_key in FOOTER_KEYWORDS:
                        if footer_key in element_text.lower():
                            is_footer = True
                            break
                    if is_footer:
                        continue

                try:
                    element_text = element_text.encode("cp1252").decode("cp1251")
                    headings = [x.encode("cp1252").decode("cp1251") for x in headings]
                except (UnicodeDecodeError, UnicodeEncodeError):
                    pass

                document_content[-1].append(element_text)
                document_meta[-1].append(
                    {
                        "type": element_meta["type"],
                        "is_heading": is_heading,
                        "is_heading_extracting_correct": correct_heading,
                        "headings": headings,
                        "paragraph": paragraph,
                    }
                )

    if remove_service_info:
        # Check if the first document's page is a title
        try:
            potential_title_page = document_content[0]
            title_num = 0
            is_title = False

            for i in range(len(document_content)):
                page = document_content[i]
                if len(page) > 0:
                    potential_title_page = page
                    title_num = i
                    break
            for text_line in potential_title_page:
                for title_key in HEADING_STOP_LIST:
                    if title_key in text_line.lower():
                        is_title = True
                        break
                if is_title:
                    break
            if not is_title:
                for char in potential_title_page[-1].lower():
                    if char.isalpha() and char != "г":
                        break
            if is_title:
                document_content = document_content[title_num + 1 :]
                document_meta = document_meta[title_num + 1 :]
        except IndexError:
            warnings.warn(
                "Can not skip title-related service information due to unknown title formatting",
                category=TitleExtractingWarning,
            )

        # Remove page numbers
        try:
            for page_number in range(len(document_content)):
                page = document_content[page_number]
                if len(page) > 0:
                    only_digits = True
                    for char in page[0]:  # number is at the beginning of the page
                        if not char.isdigit():
                            only_digits = False
                    if only_digits:
                        document_content[page_number] = document_content[page_number][
                            1:
                        ]
                        document_meta[page_number] = document_meta[page_number][1:]
                        continue
                    for char in page[-1]:  # number is at the end of the page
                        if not char.isdigit():
                            only_digits = False
                    if only_digits:
                        document_content[page_number] = document_content[page_number][
                            :-1
                        ]
                        document_meta[page_number] = document_meta[page_number][:-1]
                        continue
        except IndexError:
            warnings.warn(
                "Can not delete page numbers due to unknown formatting",
                category=PageNumbersExtractingWarning,
            )

        # # Remove footers
        # try:
        #     for page_number in range(len(document_content)):
        #         page = document_content[page_number]
        #         if len(page) > 0:
        #
        # except IndexError:
        #     warnings.warn('Can not delete page numbers due to unknown formatting',
        #                   category=FooterExtractingWarning)

    final_content = listmerge(document_content)
    final_meta = listmerge(document_meta)

    return final_content, final_meta



import logging

from langchain_core.documents import Document
from langchain_core.language_models import LLM

from protollm.templates.prompt_templates.rag_prompt_templates import PROMPT_RANK, PROMPT_LLM_RESPONSE
from protollm.rags.rag_core.reranker import LLMReranker
from protollm.rags.rag_core.retriever import DocRetriever, DocsSearcherModels, RetrievingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def get_retriever(docs_searcher_models: DocsSearcherModels, top_k: int = 5) -> DocRetriever:
    """Documents retriever object."""
    return DocRetriever(
        top_k=top_k,
        docs_searcher_models=docs_searcher_models,
    )


def run_rag(user_prompt: str, llm: LLM, retrievers: list[DocRetriever], collection_names: list[str],
            do_reranking: bool = False) -> str:
    """
    :param user_prompt: only prompt that was received from user
    :param collection_names: retriever gets docs from the collection with name 'collection_name'
    :param do_reranking: if reranking is necessary
    :return: response from LLM
    """

    # Retrieve
    logger.info('Retrieving ----------- IN PROGRESS')
    context = RetrievingPipeline() \
        .set_retrievers(retrievers) \
        .set_collection_names(collection_names) \
        .get_retrieved_docs(user_prompt)
    logger.info('Retrieving ----------- DONE')

    # Rerank
    if do_reranking:
        logger.info('Reranking ----------- IN PROGRESS')
        reranker = LLMReranker(llm, PROMPT_RANK)
        res = reranker.rerank_context(context, user_prompt)
        logger.info('Reranking ----------- DONE')
    else:
        logger.info('Reranking ----------- SKIPPED')
        res = context

    # Get response
    logger.info('Generation ----------- IN PROGRESS')
    paragraphs = "\n".join([f"Параграф {i + 1}: {doc.page_content}" for i, doc in enumerate(res)])
    llm_response = llm.invoke(PROMPT_LLM_RESPONSE.format(paragraphs=paragraphs, question=user_prompt))
    logger.info('Generation ----------- DONE')

    return llm_response


def run_multiple_rag(user_prompt: str, llm: LLM, retriever_pipelines: list[RetrievingPipeline]) -> str:
    """
    :param user_prompt: only prompt that was received from user
    :param retriever_pipelines: ready to use retriever pipelines (retrievers and collection_names should be specified)
    :param do_reranking: if reranking is necessary
    :return: response from LLM
    """
    reranker = LLMReranker(llm, PROMPT_RANK)

    # Retrieve
    logger.info('Retrieving ----------- IN PROGRESS')
    contexts = [pipeline.get_retrieved_docs(user_prompt) for pipeline in retriever_pipelines]
    logger.info('Retrieving ----------- DONE')

    max_len_context = max([len(context) for context in contexts])
    for ctx in contexts:
        if len(ctx) < max_len_context:
            ctx.extend([Document(page_content='')] * (max_len_context - len(ctx)))

    # Merge the most relevant paragraphs
    logger.info('Merging ----------- IN PROGRESS')
    context = reranker.merge_docs(user_prompt, contexts)

    # Get response
    logger.info('Generation ----------- IN PROGRESS')
    paragraphs = "\n".join([f"Параграф {i + 1}: {doc.page_content}" for i, doc in enumerate(context)])
    llm_response = llm.invoke(PROMPT_LLM_RESPONSE.format(paragraphs=paragraphs, question=user_prompt))
    logger.info('Generation ----------- DONE')

    return llm_response



import warnings
from copy import deepcopy
from typing import TextIO, Optional, Any, Type

import inspect

import yaml
from langchain_core.documents import BaseDocumentTransformer
from langchain_text_splitters import TextSplitter
from transformers import AutoTokenizer, logging

from protollm.rags.pipeline.docs_processing.entities import transformer_object_dict
from protollm.rags.pipeline.docs_processing.exceptions import TransformerNameError
from protollm.rags.pipeline.docs_processing.models import ConfigFile


def _get_params_for_transformer(params: dict[str, Any],
                                transformer_class: Type[BaseDocumentTransformer]) -> dict[str, Any]:
    text_splitter_params = inspect.signature(TextSplitter.__init__).parameters.keys()
    transformer_params = {key: value for key, value in params.items()
                          if key in inspect.signature(transformer_class.__init__).parameters.keys()
                          or key in text_splitter_params}
    return transformer_params


class PipelineSettings:
    def __init__(self, config: Optional[ConfigFile] = None):
        self.config = config
        self.loader_params = config
        self.transformers = config

    @classmethod
    def config_from_file(cls, config_file: str):
        with open(config_file, 'r') as f:
            yaml_config = yaml.safe_load(f)
        config_dict = ConfigFile.model_validate(yaml_config)
        for splitter in config_dict.splitter:
            if splitter.splitter_name not in transformer_object_dict:
                raise TransformerNameError(f'There is no DocumentTransformer '
                                           f'related to the name: {splitter.splitter_name}')
        return cls(config=config_dict)

    def update_transformer_params(self, new_params: dict[str, Any]):
        for transformer_class, transformer_params in self._transformers:
            new_param_values = _get_params_for_transformer(new_params, transformer_class)
            transformer_params.update(new_param_values)

    def update_loader_params(self, new_params: dict[str, Any]):
        self._loader_params.update(new_params)

    @property
    def loader_params(self) -> dict[str, Any]:
        return self._loader_params

    @loader_params.setter
    def loader_params(self, config: Optional[ConfigFile]):
        loader_params = deepcopy(config.loader.parsing_params)
        loader_params['file_path'] = config.loader.file_path
        self._loader_params = loader_params

    @property
    def config(self) -> ConfigFile:
        return self._config

    @config.setter
    def config(self, config: Optional[ConfigFile]):
        self._config = deepcopy(config)

    @property
    def transformers(self) -> list[BaseDocumentTransformer]:
        docs_transformers = []
        tokenizer = None
        if self.config.tokenizer is not None:
            logging.set_verbosity_error()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                tokenizer = AutoTokenizer.from_pretrained(self.config.tokenizer)
        for transformer_class, transformer_params in self._transformers:
            if tokenizer is None or not hasattr(transformer_class, 'from_huggingface_tokenizer'):
                transformer = transformer_class(**transformer_params)
            else:
                transformer = transformer_class.from_huggingface_tokenizer(
                    tokenizer,
                    **transformer_params
                )
            docs_transformers.append(transformer)
        return docs_transformers

    @transformers.setter
    def transformers(self, config: Optional[ConfigFile]):
        self._transformers = []
        if config is None:
            return
        self._transformers.append((transformer_object_dict['recursive_character'], {}))
        for splitter in config.splitter:
            splitter_params = splitter.splitter_params
            transformer_class = transformer_object_dict[splitter.splitter_name]
            transformer_params = _get_params_for_transformer(splitter_params, transformer_class)
            if splitter.splitter_name != 'recursive_character':
                self._transformers.append((transformer_class, transformer_params))
            else:
                self._transformers[0] = (transformer_class, transformer_params)



import re
import warnings
from pathlib import Path
from typing import Iterator, Union

from langchain_core.document_loaders import Blob
from langchain_core.documents import Document

from protollm.raw_data_processing.docs_parsers.utils.exceptions import (
    EncodingError,
    ChaptersExtractingFailedWarning,
)
from protollm.raw_data_processing.docs_parsers.parsers.base import BaseParser
from protollm.raw_data_processing.docs_parsers.parsers.entities import ParsingScheme
from protollm.raw_data_processing.docs_parsers.parsers.utilities import CONTENTS_KEYWORDS
from protollm.raw_data_processing.docs_parsers.utils.utilities import correct_path_encoding, is_bad_encoding


class PDFParser(BaseParser):
    """
    The parser provides a way to parse raw data from PDF into one or more documents.
    """

    def __init__(
        self,
        parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        extract_images: bool = False,
        extract_tables: bool = False,
        parse_formulas: bool = False,
        remove_service_info: bool = False,
    ):
        try:
            import protollm.raw_data_processing.docs_parsers.parsers.pdf.utilities
        except ImportError as error:
            raise ImportError(
                f"{error.name} package not found, please try to install it with `pip install {error.name}`"
            )
        if parsing_scheme not in ParsingScheme.__members__:
            raise ValueError("Invalid parsing scheme")
        self.parsing_scheme = parsing_scheme
        self.extract_images = extract_images
        self.extract_tables = extract_tables
        self.parse_formulas = parse_formulas
        self.remove_service_info = remove_service_info

    def lazy_parse(self, blob: Blob) -> Iterator[Document]:
        from protollm.raw_data_processing.docs_parsers.parsers.pdf.utilities import extract_by_lines

        source = blob.source
        source = correct_path_encoding(source) if source is not None else ""
        file_name = Path(source).name

        with blob.as_bytes_io() as pdf_file_obj:
            lines, metadata = extract_by_lines(
                pdf_file_obj,
                parse_images=self.extract_images,
                parse_tables=self.extract_tables,
                parse_formulas=self.parse_formulas,
                remove_service_info=self.remove_service_info,
            )

        if not lines:
            return

        if is_bad_encoding(lines):
            raise EncodingError(
                "It is impossible to parse the file due to uncertainty in the text encoding"
            )

        # Get info about maximum titles hierarchy level
        max_hierarchy_lvl = max([len(x["headings"]) for x in metadata])

        is_heading_extracting_correct = True
        for meta in metadata:
            if not meta["is_heading_extracting_correct"]:
                is_heading_extracting_correct = False
                break

        if max_hierarchy_lvl != 0:
            upd_lines = []
            upd_metadata = []
            for text, meta in zip(lines, metadata):
                if self.remove_service_info:
                    if (
                        len(meta["headings"]) > 0
                        and meta["headings"][0].lower() in CONTENTS_KEYWORDS
                    ):  # len(meta['headings']) == 0 or
                        continue
                upd_lines.append(text)
                upd_metadata.append(meta)
            lines = upd_lines
            metadata = upd_metadata
        else:
            for i in range(len(metadata)):
                metadata[i]["headings"].append("Документ")

        match self.parsing_scheme:
            case ParsingScheme.lines:
                for text, meta in zip(lines, metadata):
                    yield Document(
                        page_content=text,
                        metadata={**meta, "source": source, "file_name": file_name},
                    )
            case ParsingScheme.full:
                text = " ".join([x.strip() for x in lines])
                pattern = r"(?<=[А-Яа-яёЁ])-\s"
                text = re.sub(pattern, "", text)
                meta = {
                    "page": "all",
                    "source": source,
                    "file_name": file_name,
                    "is_heading_extracting_correct": is_heading_extracting_correct,
                }
                yield Document(page_content=text, metadata=meta)
            case ParsingScheme.chapters:
                if not is_heading_extracting_correct:
                    warnings.warn(
                        "Can not correctly extract chapters due to complex document structure",
                        category=ChaptersExtractingFailedWarning,
                    )
                text_lst = []
                heading = ""
                for text, meta in zip(lines, metadata):
                    if len(meta["headings"]) != 0:  # text is a part of some chapter
                        if (
                            meta["headings"][0] == heading
                        ):  # it has the same chapter's heading
                            text_lst.append(
                                text
                            )  # add element's text to the document's content
                        elif heading != "":  # other heading
                            document_text = " ".join(text_lst)
                            document_meta = {
                                "heading": heading,
                                "source": source,
                                "file_name": file_name,
                                "is_heading_extracting_correct": is_heading_extracting_correct,
                            }

                            text_lst = [text]
                            heading = meta["headings"][0]
                            pattern = r"(?<=[А-Яа-яёЁ])-\s"
                            document_text = re.sub(pattern, "", document_text)
                            yield Document(
                                page_content=document_text, metadata=document_meta
                            )
                        else:
                            text_lst = [text]
                            heading = meta["headings"][0]
                if len(text_lst) != 0:
                    document_text = " ".join(text_lst)
                    pattern = r"(?<=[А-Яа-яёЁ])-\s"
                    document_text = re.sub(pattern, "", document_text)
                    document_meta = {
                        "heading": heading,
                        "source": source,
                        "file_name": file_name,
                        "is_heading_extracting_correct": is_heading_extracting_correct,
                    }

                    yield Document(page_content=document_text, metadata=document_meta)
            case ParsingScheme.paragraphs:
                text_lst = []
                paragraph = -1
                document_meta = {
                    "source": source,
                    "file_name": file_name,
                    "is_heading_extracting_correct": is_heading_extracting_correct,
                }
                for text, meta in zip(lines, metadata):
                    if len(meta["headings"]) != 0:  # text is a part of some chapter
                        if (
                            meta["paragraph"] != -1
                        ):  # text is a part of some paragraph (not header)
                            if (
                                meta["paragraph"] == paragraph
                            ):  # it has the same paragraph's number
                                text_lst.append(
                                    text
                                )  # add element's text to the document's content
                                document_meta = {
                                    **meta,
                                    "source": source,
                                    "file_name": file_name,
                                    "is_heading_extracting_correct": is_heading_extracting_correct,
                                }
                            elif paragraph != -1:  # other paragraph
                                document_text = " ".join(text_lst)
                                text_lst = [text]
                                paragraph = meta["paragraph"]
                                pattern = r"(?<=[А-Яа-яёЁ])-\s"
                                document_text = re.sub(pattern, "", document_text)
                                yield Document(
                                    page_content=document_text, metadata=document_meta
                                )
                            else:
                                text_lst = [text]
                                paragraph = meta["paragraph"]
                if len(text_lst) != 0:
                    document_text = " ".join(text_lst)
                    pattern = r"(?<=[А-Яа-яёЁ])-\s"
                    document_text = re.sub(pattern, "", document_text)
                    yield Document(page_content=document_text, metadata=document_meta)
            case _:
                raise NotImplementedError(
                    f"{self.parsing_scheme} type of parsing scheme is not implemented"
                )



import re

# HEADINGS

HEADING_KEYWORDS = [
    "предисловие",
    "содержание",
    "оглавление",
    "введение",
    "лист согласований",
]
CONTENTS_KEYWORDS = ["содержание", "оглавление", "лист согласований"]
HEADING_STOP_LIST = ["утверждаю"]

FOOTER_KEYWORDS = ["документ создан в электронной форме", "страница создана"]


# BULLETS

UNICODE_BULLETS = [
    "\u0095",
    "\u2022",
    "\u2023",
    "\u2043",
    "\u3164",
    "\u204C",
    "\u204D",
    "\u2219",
    "\u25CB",
    "\u25CF",
    "\u25D8",
    "\u25E6",
    "\u2619",
    "\u2765",
    "\u2767",
    "\u29BE",
    "\u29BF",
    "\u002D",
    "",
    r"\*",
    "\x95",
    "·",
]

BULLETS_PATTERN = "|".join(UNICODE_BULLETS)
UNICODE_BULLETS_RE = re.compile(f"(?:{BULLETS_PATTERN})(?!{BULLETS_PATTERN})")


def is_bulleted_text(text: str) -> bool:
    """Checks to see if the section of text is part of a bulleted list."""
    return UNICODE_BULLETS_RE.match(text.strip()) is not None



import logging
from typing import Any, Iterable, List

from langchain_core.documents import Document
from langchain_text_splitters import TextSplitter

logger = logging.getLogger(__name__)


def _get_metadata(metadata: dict[str, Any]) -> dict[str, Any]:
    chapter = metadata["headings"][0] if metadata.get("headings", []) else ""
    source = metadata.get("source", "")
    meta = {
        "chapter": chapter,
        "source": source,
        "file_name": metadata.get("file_name", ""),
        "keywords": metadata.get("keywords", []),
        "object": [metadata.get("object", "")],
        "action": [metadata.get("action", "")],
    }
    return meta


class ChunkMerger(TextSplitter):
    def __init__(self, joiner: str = " ", **kwargs: Any):
        super().__init__(**kwargs)
        self.joiner = joiner

    def _merge_documents(self, documents: Iterable[Document]) -> List[Document]:
        documents = list(documents)
        if not documents:
            return []
        first_doc = documents[0]
        merged_chunk_content = first_doc.page_content
        merged_chunk_meta = _get_metadata(first_doc.metadata)
        chapter, source = merged_chunk_meta["chapter"], merged_chunk_meta["source"]
        transformed_docs = []
        for doc in documents[1:]:
            doc_text = doc.page_content
            doc_meta = _get_metadata(doc.metadata)
            doc_chapter, doc_source = doc_meta["chapter"], doc_meta["source"]

            if (
                doc_chapter == chapter and doc_source == source
            ):
                merged_text = self.joiner.join((merged_chunk_content, doc_text))
                if self._length_function(merged_text) <= self._chunk_size:
                    merged_chunk_content = merged_text
                    merged_chunk_meta["keywords"] += doc_meta["keywords"]
                    merged_chunk_meta["object"] += doc_meta["object"]
                    merged_chunk_meta["action"] += doc_meta["action"]
                    continue

            if merged_chunk_content:
                merged_chunk_meta["keywords"] = ", ".join(
                    set(merged_chunk_meta["keywords"])
                )
                merged_chunk_meta["object"] = ", ".join(
                    set(merged_chunk_meta["object"])
                )
                merged_chunk_meta["action"] = ", ".join(
                    set(merged_chunk_meta["action"])
                )
                transformed_docs.append(
                    self._create_document(merged_chunk_content, merged_chunk_meta)
                )

            merged_chunk_content, merged_chunk_meta = doc_text, doc_meta
            chapter, source = doc_chapter, doc_source

        if merged_chunk_content:
            transformed_docs.append(
                self._create_document(merged_chunk_content, merged_chunk_meta)
            )

        return transformed_docs

    def _create_document(self, text: str, metadata: dict[str, Any]) -> Document:
        text_len = self._length_function(text)
        if text_len > self._chunk_size:
            logger.warning(
                f"A chunk of size {text_len} was encountered, "
                f"which is larger than the specified {self._chunk_size}"
            )
        return Document(page_content=text, metadata=metadata)

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        return self._merge_documents(documents)

    def split_text(self, text: str) -> List[str]:
        raise NotImplementedError("split_text is not implemented in ChunkMerger")

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        raise NotImplementedError(
            "_merge_splits is not implemented in ChunkMerger"
        )



import logging
import os
from typing import Any, Iterable, Optional, List, Tuple

import spacy
from langchain_core.documents import Document
from langchain_text_splitters import TextSplitter

logger = logging.getLogger(__name__)


class MultiMetadataAppender(TextSplitter):
    def __init__(self, separators: Optional[Iterable[str]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._separators = separators or [r"\. (?<![0-9]\. | ^[0-9].)"]
        self.keyword_extractor = KeywordExtractor()

    def _create_document(self, text: str, metadata: dict[str, Any]) -> Document:
        text_len = self._length_function(text)
        if text_len > self._chunk_size:
            logger.warning(
                f"A chunk of size {text_len} was encountered, "
                f"which is larger than the specified {self._chunk_size}"
            )
        return Document(page_content=text, metadata=metadata)

    def _split_with_additional_metadata(
        self, documents: Iterable[Document]
    ) -> List[Document]:
        texts, metadatas = [], []
        for doc in documents:
            doc_metadata = doc.metadata
            doc_metadata["object"], doc_metadata["action"] = (
                self.keyword_extractor.get_object_action_pair(doc.page_content)
            )
            doc_metadata["keywords"] = self.keyword_extractor.get_keywords(
                doc.page_content
            )
            texts.append(doc.page_content)
            metadatas.append(doc_metadata)
        return self.create_documents(texts, metadatas=metadatas)

    def split_text(self, text: str) -> List[str]:
        return [text]

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        return self._split_with_additional_metadata(documents)


class KeywordExtractor:
    def __init__(self):
        os.system("python -m spacy download ru_core_news_sm")
        self.nlp = spacy.load("ru_core_news_sm")

    def get_keywords(self, text: str) -> List[str]:
        tokens = self.nlp(text)
        subjects = []
        for token in tokens:
            if "nsubj" in token.dep_ or "obj" in token.dep_:
                subjects.append(token)
                break

        result_tokens = subjects.copy()

        for token in subjects:
            children = [tk for tk in token.children if tk.dep_ == "nmod"]
            result_tokens.extend(children)
            while len(children) > 0:
                ch = children.pop()
                tmp = [child for child in ch.children if child.dep_ == "nmod"]
                result_tokens.extend(tmp)
                children.extend(tmp)

        return list(set([token.lemma_ for token in result_tokens]))

    def get_object_action_pair(self, text: str) -> Tuple[str, str]:
        tokens = self.nlp(text)
        for token in tokens:
            if "nsubj" in token.dep_ or "obj" in token.dep_:
                return token.lemma_, token.head.lemma_
        return "", ""



import logging
import warnings
from contextlib import contextmanager
from typing import Optional, Generator


class ParsingLogger:
    def __init__(self, silent_errors: bool = False, name: Optional[str] = None):
        name = name or __name__
        self._logger = logging.getLogger(name)
        self._logs: dict[str, list[str]] = {}
        self._silent_errors = silent_errors

    @property
    def logger(self):
        return self._logger

    @property
    def logs(self):
        return self._logs

    def info(self, msg: str, *args, **kwargs):
        self._logger.info(msg, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        self._logger.warning(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        self._logger.error(msg, *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs):
        self._logger.critical(msg, *args, **kwargs)

    def exception(self, msg: str, *args, **kwargs):
        self._logger.exception(msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        self._logger.debug(msg, *args, **kwargs)

    @contextmanager
    def parsing_info_handler(self, file_name: str) -> Generator[None, None, None]:
        try:
            try:
                with warnings.catch_warnings(record=True) as record:
                    warnings.simplefilter("default")
                    yield
            finally:
                for warning in record:
                    warn_msg = f"{warning.message} (in {file_name})"
                    self.warning(warn_msg)
        except Exception as error:
            self.logs[file_name] = self.logs.get(file_name, [])
            self.logs[file_name].append(str(error))
            err_msg = f"{error} (in {file_name})"
            self.error(err_msg)
            if not self._silent_errors:
                raise



import logging
import re
from typing import Any, Iterable, Optional, List

from langchain_core.documents import Document
from langchain_text_splitters import TextSplitter

from protollm.raw_data_processing.docs_transformers.utilities import fix_list_dots_separators

logger = logging.getLogger(__name__)


class DivMetadataSentencesSplitter(TextSplitter):
    """
    Splitter that divide document into sentences and add the source of this sentence to metadata.

    Example:
        Document:
        - page_content = "There is a huge number of clustering algorithms. The main idea of most of them is to combine
            identical sequences into one class or cluster based on similarity.
            As a rule, the choice of algorithm is determined by the task at hand."
        - metadata = {...}

        After splitting there are 3 Documents:
        [
            Document_1:
                - page_content = "There is a huge number of clustering algorithms."
                - metadata = {the same of Document, 'div': 'There is a huge number of clustering algorithms.
                    The main idea of most of them is to combine identical sequences into one class or cluster based on similarity.
                    As a rule, the choice of algorithm is determined by the task at hand.'}
            Document_2:
                - page_content = "The main idea of most of them is to combine identical sequences into one class or cluster based on similarity."
                - metadata = {the same of Document, 'div': 'There is a huge number of clustering algorithms.
                    The main idea of most of them is to combine identical sequences into one class or cluster based on similarity.
                    As a rule, the choice of algorithm is determined by the task at hand.'}
            Document_3:
                - page_content = "As a rule, the choice of algorithm is determined by the task at hand.",
                - metadata = {the same of Document, 'div': 'There is a huge number of clustering algorithms.
                    The main idea of most of them is to combine identical sequences into one class or cluster based on similarity.
                    As a rule, the choice of algorithm is determined by the task at hand.'}
        ]
    """

    def __init__(self, separators: Optional[Iterable[str]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._separators = separators or [r"\. (?<![0-9]\. | ^[0-9].)"]

    def _create_document(self, text: str, metadata: dict[str, Any]) -> Document:
        text_len = self._length_function(text)
        if text_len > self._chunk_size:
            logger.warning(
                f"A chunk of size {text_len} was encountered, "
                f"which is larger than the specified {self._chunk_size}"
            )
        return Document(page_content=text, metadata=metadata)

    def _split_on_sentences_with_additional_metadata(
        self, documents: Iterable[Document]
    ) -> List[Document]:
        splitted_docs = []
        for doc in documents:
            doc_metadata = doc.metadata
            doc_metadata["div"] = doc.page_content
            splitted_docs.extend(
                [
                    self._create_document(text, doc_metadata)
                    for text in self.split_text(doc.page_content)
                ]
            )

        return splitted_docs

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        return self._split_on_sentences_with_additional_metadata(documents)

    def split_text(self, text: str) -> List[str]:
        sentences_lst = [
            x.strip() for x in re.split(" | ".join(self._separators), text)
        ]
        return fix_list_dots_separators(sentences_lst)



import warnings
from typing import List, Tuple

from langchain_core.documents import Document
from langchain_core.language_models import LLM
from langchain_core.prompts import PromptTemplate


class LLMReranker:
    def __init__(self, llm: LLM, prompt_template: PromptTemplate):
        """
        Reranker to change the order of documents using LLM.

        :param prompt_template: prompt template for reranking. It should contain the 'question' and 'context' fields
        """
        self._prompt_template = prompt_template
        # The retries number if the first model response was obtained in wrong format or have another anomalies
        # that are not characteristic for correct operation of the model in accordance with the prompt template
        self.num_retries = 3
        # The lower boundary of context LLM estimation
        self.qual_threshold = 2
        self._llm = llm

    def rerank_context(self, context: list[Document], user_query: str, top_k: int = 3) -> list[Document]:
        ranking_prompts = [self._prompt_template.format(question=user_query,
                                                        context=context_i.page_content +
                                                                " Имя файла, откуда взят параграф " +
                                                                context_i.metadata.get('source',
                                                                                       '/None').split('/')[-1])
                           for context_i in context]
        answers_ranking, bad_query = self._get_ranking_answer(ranking_prompts)
        if bad_query:
            fixed_answers = self._regenerate_answer(bad_query)
            answers_ranking += fixed_answers
        ext_context = self._extract_top_context(answers_ranking, top_k)
        if not ext_context:
            warnings.warn('Reranker does not support retrieved context')
        res_context = [context[ranking_prompts.index(i)] for i in ext_context]
        return res_context

    def _extract_top_context(self, pairs_to_rank: List[Tuple[str, int]], top_k: int) -> list[str]:
        if not pairs_to_rank:
            return []
        pairs_to_rank.sort(key=lambda x: x[1], reverse=True)
        context = [x for x, y in pairs_to_rank if y >= self.qual_threshold]
        context = context[:top_k]
        return context

    def _get_ranking_answer(self, ranking_prompts: list[str]) -> Tuple[list[Tuple[str, int]], list[str]]:
        answer = [self._llm.invoke(prompt) for prompt in ranking_prompts]
        answers_ranking = []
        bad_queries = []
        for i, ans_i in enumerate(answer):
            try:
                score = int(ans_i.split('ОЦЕНКА: ')[-1].strip())
                answers_ranking.append((ranking_prompts[i], score))
            except:
                bad_queries.append(ranking_prompts[i])
        return answers_ranking, bad_queries

    def _regenerate_answer(self, queries: list[str]) -> list[str]:
        fixed_queries = []
        for i in range(self.num_retries):
            good_res, bad_res = self._get_ranking_answer(queries)
            fixed_queries += good_res
            queries = bad_res
            if not bad_res:
                return fixed_queries
        return fixed_queries

    def merge_docs(self, query: str, contexts: list[list[Document]], top_k: int = 3) -> list[Document]:
        ctx = []
        for context in zip(*contexts):
            ctx.extend(self.rerank_context(context, query, 1))

        if len(ctx) > top_k:
            return self.rerank_context(ctx, query, top_k)

        return ctx




from functools import partial
from pathlib import Path
from typing import Iterator, Union, Optional

from langchain_core.document_loaders import Blob
from langchain_core.documents import Document

from protollm.raw_data_processing.docs_parsers.utils.exceptions import EncodingError
from protollm.raw_data_processing.docs_parsers.parsers.base import BaseParser
from protollm.raw_data_processing.docs_parsers.parsers.entities import ParsingScheme
from protollm.raw_data_processing.docs_parsers.parsers.word_doc.utilities import (
    get_paragraphs,
    get_chapters,
    add_headings_hierarchy,
)
from protollm.raw_data_processing.docs_parsers.utils.utilities import correct_path_encoding, is_bad_encoding


class WordDocumentParser(BaseParser):
    """
    The parser provides a way to parse raw data from Word Document into one or more documents.
    """

    def __init__(
        self,
        parsing_scheme: Union[ParsingScheme, str] = ParsingScheme.lines,
        extract_images: bool = False,
        extract_tables: bool = False,
        extract_formulas: bool = False,
        timeout_for_converting: Optional[int] = None,
    ):
        try:
            import protollm.raw_data_processing.docs_parsers.parsers.word_doc.docx_parsing
        except Exception as e:
            raise e
        # except ImportError as error:
        #     raise ImportError(
        #         f"{error.name} package not found, please try to install it with `pip install {error.name}`"
        #     )
        if parsing_scheme not in ParsingScheme.__members__:
            raise ValueError("Invalid parsing scheme")
        self.parsing_scheme = parsing_scheme
        self.extract_images = extract_images
        self.extract_tables = extract_tables
        self.extract_formulas = extract_formulas
        self.timeout = timeout_for_converting

    def lazy_parse(self, blob: Blob) -> Iterator[Document]:
        from protollm.raw_data_processing.docs_parsers.parsers.word_doc.docx_parsing import (
            parse_docx_to_lines,
        )

        parse_docx_to_lines = partial(
            parse_docx_to_lines,
            extract_tables=self.extract_tables,
            extract_images=self.extract_images,
            extract_formulas=self.extract_formulas,
        )

        match blob.mimetype:
            case "doc" | "odt" | "rtf":
                from protollm.raw_data_processing.docs_parsers.parsers.converting import (
                    converted_file_to_docx,
                )

                with blob.as_bytes_io() as file_obj:
                    with converted_file_to_docx(
                        file_obj, timeout=self.timeout
                    ) as docx_file_obj:
                        lines, metadata = parse_docx_to_lines(docx_file_obj)
            case "docx":
                with blob.as_bytes_io() as docx_file_obj:
                    lines, metadata = parse_docx_to_lines(docx_file_obj)
            case _:
                raise ValueError("Invalid document type")

        if is_bad_encoding(lines):
            raise EncodingError(
                "It is impossible to parse the file due to uncertainty in the text encoding"
            )

        source = blob.source
        source = correct_path_encoding(source) if source is not None else ""
        file_name = Path(source).name

        if self.parsing_scheme == ParsingScheme.full:
            text = " ".join(lines)
            meta = {
                "page": "all",
                "headings": [],
                "source": source,
                "file_name": file_name,
            }
            yield Document(page_content=text, metadata=meta)
            return

        lines, metadata = add_headings_hierarchy(lines, metadata)

        match self.parsing_scheme:
            case ParsingScheme.lines:
                texts = lines
            case ParsingScheme.paragraphs:
                texts, metadata = get_paragraphs(lines, metadata)
            case ParsingScheme.chapters:
                texts, metadata = get_chapters(lines, metadata)
            case _:
                raise NotImplementedError(
                    f"{self.parsing_scheme} type of parsing scheme is not implemented"
                )

        for text, meta in zip(texts, metadata):
            yield Document(
                page_content=text,
                metadata={**meta, "source": source, "file_name": file_name},
            )



from langchain_core.prompts import PromptTemplate

PROMPT_RANK_TEMPLATE = """
<|begin_of_text|>
<|start_header_id|>system<|end_header_id|>
Ты — Сайга, русскоязычный автоматический ассистент. Ты разговариваешь с людьми и помогаешь им.

Ты - система осуществляющая отбор текстов подходящих к запросу для дальнейшей передаче информации системе генерации финального ответа на запрос. \
Тебе будет предоставлен ЗАПРОС пользователя и ТЕКСТ. Твоя задача в 10-ти балльной шкале оценить, насколько ТЕКСТ несет полезную информацию \
для ЗАПРОСА, т.е. можно ли ответить на ЗАПРОС, используя информацию из ТЕКСТА. Нет требования, что ЗАПРОС должен быть \
удовлетворен только информацией из этого ТЕКСТА может быть еще будут ТЕКСТЫ, которые дополнят ответ.

Пользуйся следующим набором правил:
1. В первую очередь нужно вывести рассуждение соответствует ли текст запросу, в чем именно соответствует и почему
2. Только после рассуждения выводи оценку цифрой

Сначала напиши объяснение оценки, а затем выведи саму оценку "ОЦЕНКА: твоя оценка" 
<|eot_id|>
<|start_header_id|>user<|end_header_id|>

ЗАПРОС:
{question}

ТЕКСТ: 
{context}

<|eot_id|>
<|start_header_id|>assistant<|end_header_id|>
"""

# TODO: remake it for more high accuracy
PROMPT_PLAN_TEMPLATE = """
<|begin_of_text|>
<|start_header_id|>system<|end_header_id|>
Ты система осуществляющая планирование запросов нужной для пользователя информации в векторной базе с документами разделенными на небольшие сегменты. \
Для каждого поступившего запроса от пользователя \
твоя задача, при необходимости, разделить его на несколько запросов к векторной базе данных для получения нужной для ответа информации. \
Ты должен написать итоговые запросы "Запросы:", а потом написать объяснение "Объяснение:". И запросы и объяснение должны быть на русском языке. \
Например, вопрос "назови столицу России" не нуждается в разделении, необходимая информация о столице России будет извлечена. \
В свою очередь, запрос "Сравни автомобили ауди q5 и audi q7" уже должен быть разделен на два: "Какие преимущества и недостатки и audi q5" и "Какие преимущества и недостатки у audi q7", т.к. нам \
нужна информация по обоим объектам для проведения сравнения. 
Пользуйся следующим набором правил:
1. Если встречается указание на год или день, то учти, что сейчас 14 июня 2024 года и заменяй в итоговых запросах на конкретную информацию для улучшения поиска (например, "в этом году" должно быть изменено на "2024 году", "сегодня" нужно заменить на "14 июня 2024 года"). 
2. Сфера для которой ты работаешь и в которой работают пользователи - это строительство.
3. Каждый следующий вопрос не должен опираться на результат ответа предыдущего (нельзя: "Какие из этих технологий..."). Агрегация информации произойдет позднее.
4. Если будешь отвечать на английском получишь штраф.
5. Учитывай что у пользовательского запроса есть контекст, который нельзя терять.
6. Если пользовательский запрос можно разделить на несколько, не забудь добавить необходимый контекст для независимого поиска по базе.

Результат выведи списком запросов после ключевого слова ЗАПРОСЫ: ["запрос1", "запрос2"...]

<|eot_id|>
<|start_header_id|>user<|end_header_id|>

{context}

<|eot_id|>
<|start_header_id|>assistant<|end_header_id|>
"""

PROMPT_LLM_RESPONSE_TEMPLATE = """
<|begin_of_text|>
<|start_header_id|>system<|end_header_id|>
Ты эксперт в области строительства, который помогает новичкам ответить на вопросы по технологии. Для этого у тебя есть экспертная информация из документов, которая представлена в виде параграфов. Ответь на вопрос, выдавая ВСЮ ИНФОРМАЦИЮ из параграфа. Для ответа используй только информацию в представленных параграфах.
{paragraphs}
<|eot_id|>
<|start_header_id|>user<|end_header_id|>

{question}

<|eot_id|>
<|start_header_id|>assistant<|end_header_id|>
"""

PROMPT_REFORMAT_TEMPLATE = """
<|begin_of_text|>
<|start_header_id|>system<|end_header_id|>
Тебе дан вопрос, который получен от пользователя. Говори на русском языке. 
Твоя задача - изменить этот вопрос, не отвечая на него, оставив только главную суть предложения.
<|eot_id|>
<|start_header_id|>user<|end_header_id|>

{question}

<|eot_id|>
<|start_header_id|>assistant<|end_header_id|>
"""

PROMPT_REFORMAT = PromptTemplate(template=PROMPT_REFORMAT_TEMPLATE, input_variables=['question'])
PROMPT_PLAN = PromptTemplate(template=PROMPT_PLAN_TEMPLATE, input_variables=['context'])
PROMPT_RANK = PromptTemplate(template=PROMPT_RANK_TEMPLATE, input_variables=['question', 'context'])
PROMPT_LLM_RESPONSE = PromptTemplate(template=PROMPT_LLM_RESPONSE_TEMPLATE, input_variables=['paragraphs', 'question'])

BOS, EOS = "<|begin_of_text|>", "<|end_of_text|>"
SOH, EOH = "<|start_header_id|>", "<|end_header_id|>\n\n"
EOT = "<|eot_id|>"



import json

import requests
from langchain.agents import create_structured_chat_agent, AgentExecutor
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.prompts import SystemMessagePromptTemplate, HumanMessagePromptTemplate
from langchain_core.tools import tool

from protollm.agents.llama31_agents.llama31_agent import Llama31ChatModel


@tool
def tool_example(var1: int) -> dict:
    """
    External tool.
    """
    params = {
        "var1": var1,
    }
    resp = requests.post('external_url', data=json.dumps(params))

    ans = json.loads(resp.json())

    return ans


tools = [tool_example]

# Create the system and human prompts
system_prompt = '''
Respond to the human as helpfully and accurately as possible. You have access to the following tools:

{tools}

Use a JSON blob to specify a tool by providing an "action" key (tool name) and an "action_input" key (tool input).

Valid "action" values: "Final Answer" or {tool_names}

Provide only ONE action per JSON blob, as shown:

{{ "action": $TOOL_NAME, "action_input": $INPUT }}

Follow this format:

Question: input question to answer
Thought: consider previous and subsequent steps
Action: $JSON_BLOB

Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action: {{ "action": "Final Answer", "action_input": "Final response to human" }}


Begin! Reminder to ALWAYS respond with a valid JSON blob of a single action. Use tools if necessary. 
Respond directly if appropriate. Format is Action:```$JSON_BLOB``` then Observation
In the "Final Answer" you must ALWAYS do required action!!!
For example answer must consist table (!):
<table examples here>
'''

human_prompt = '''{input}
{agent_scratchpad}
(Reminder to respond in a JSON blob no matter what)'''

system_message = SystemMessagePromptTemplate.from_template(
    system_prompt,
    input_variables=["tools", "tool_names"],
)
human_message = HumanMessagePromptTemplate.from_template(
    human_prompt,
    input_variables=["input", "agent_scratchpad"],
)

# Create the ChatPromptTemplate
prompt = ChatPromptTemplate.from_messages(
    [
        system_message,
        MessagesPlaceholder(variable_name="chat_history", optional=True),
        human_message,
    ]
)

# Initialize the custom LLM
llm = Llama31ChatModel(
    api_key='API_KEY_HERE',
    base_url="URL_HERE",
    model="meta-llama/llama-3.1-70b-instruct",
    temperature=0.5,
    max_tokens=5000
)

# Create the structured chat agent
agent = create_structured_chat_agent(
    llm=llm,
    tools=tools,
    prompt=prompt,
    stop_sequence=True
)

# Create the AgentExecutor
agent_executor = AgentExecutor.from_agent_and_tools(
    agent=agent,
    tools=tools,
    verbose=True,
    return_intermediate_steps=True,
    output_keys=["output"],
    early_stopping_method="generate"
)

response = agent_executor.invoke({
    "input": 'test'
})



import io
from uuid import uuid4

from PIL import Image
from docx.text.paragraph import Paragraph
from lxml import etree

from protollm.raw_data_processing.docs_parsers.parsers.word_doc.xml.xml_tag import XMLTag


def _convert_to_latex(
    xml_element: etree.Element, parsing_config: 'DocxParsingConfig'
) -> str:
    math_ml = parsing_config.omml2mml_transformation(xml_element).getroot()
    tex = parsing_config.mml2tex_transformation(math_ml)
    return str(tex)


def _extract_image_data(
    xml_element: etree.Element, parsing_config: 'DocxParsingConfig'
) -> dict:
    rid = xml_element.find(".//" + XMLTag.blip.value).get(XMLTag.embed.value)
    image_element = parsing_config.document_relationships[rid].target_part
    image = Image.open(io.BytesIO(image_element.blob))
    # [Example Future Functionality] extracted_data = process_image(image)  # TODO: add image processing
    return {"image_filename": image_element.filename, "image": image}


def _parse_raw_xml_element(
    raw_xml_element: etree.Element, parsing_config: 'DocxParsingConfig'
) -> tuple[list[str], dict]:
    """Parse raw xml element."""
    texts = []
    extracted_image_data = {}

    for child_element in raw_xml_element:
        match child_element.tag:
            case XMLTag.text:
                texts.append(child_element.text)
            case XMLTag.raw:
                child_texts, child_extracted_image_data = _parse_raw_xml_element(
                    child_element, parsing_config
                )
                texts.extend(child_texts)
                extracted_image_data.update(child_extracted_image_data)
            case XMLTag.image:
                if parsing_config.extract_images:
                    image_name = "image-" + str(uuid4())
                    texts.append(f"{{{image_name}}}")
                    extracted_image_data[image_name] = _extract_image_data(
                        child_element, parsing_config
                    )

    return texts, extracted_image_data


def process_paragraph_body(
    paragraph: Paragraph, parsing_config: 'DocxParsingConfig'
) -> tuple[str, dict[str, dict]]:
    xml_paragraph = etree.fromstring(paragraph._element.xml)
    texts = []
    extracted_data = {"images": {}, "formulas": {}}
    for element in xml_paragraph:
        match element.tag:
            case XMLTag.raw:
                extracted_texts, extracted_image_data = _parse_raw_xml_element(
                    element, parsing_config
                )
                texts.extend(extracted_texts)
                extracted_data["images"].update(extracted_image_data)
            case XMLTag.math | XMLTag.math_paragraph:
                if parsing_config.parse_formulas:
                    formula_name = "formula-" + str(uuid4())
                    texts.append(f"{{{formula_name}}}")
                    extracted_data["formulas"][formula_name] = _convert_to_latex(
                        element, parsing_config
                    )

    return "".join(texts), extracted_data
